from typing import List, Optional

from yardstick.artifact import LabelEntry, Match

# note: this cannot be changed, as there is persisted values with this value + store logic based on this
MANUAL_SOURCE = "manual"


def label_entry_matches_image_lineage(
    label_entry: LabelEntry,
    image: Optional[str],
    lineage: Optional[List[str]] = None,
):
    if not lineage:
        lineage = []
    return any(label_entry.matches_image(i) for i in [image, *lineage])


def find_labels_for_match(  # noqa: PLR0913, PLR0912, C901
    image: Optional[str],
    match: Match,
    label_entries: List[LabelEntry],
    lineage: Optional[List[str]] = None,
    must_match_image=True,
    fuzzy_package_match=False,
) -> List[LabelEntry]:
    matched_label_entries: List[LabelEntry] = []
    for label_entry in label_entries:
        # this field must be matched to continue
        if not has_overlapping_vulnerability_id(label_entry, match):
            continue

        # this field must be matched to continue
        if must_match_image and not label_entry_matches_image_lineage(
            label_entry,
            image,
            lineage,
        ):
            continue

        # we need at least one more field to match to consider this label valid for the given match...
        matched_fields = 0

        if label_entry.package:
            if label_entry.package != match.package:
                # if fuzzy mathcing isn't enabled, bail now
                if not fuzzy_package_match:
                    continue

                # names must minimally match with some normalization
                if label_entry.package.name.replace(
                    "-",
                    "_",
                ) != match.package.name.replace("-", "_"):
                    continue

                # version must be a subset or superset of other
                if (
                    label_entry.package.version != match.package.version
                    and label_entry.package.version not in match.package.version
                    and match.package.version not in label_entry.package.version
                ):
                    continue

            matched_fields += 1

        if label_entry.fullentry_fields:
            mismatched = False
            for value in label_entry.fullentry_fields:
                if not _contains_as_value(match.fullentry, value):
                    mismatched = True
                    break
            if mismatched:
                continue
            matched_fields += 1

        if matched_fields > 0:
            # we should match on a minimum number of fields, otherwise a blank entry with a vuln ID will match, which is wrong
            matched_label_entries.append(label_entry)
    return matched_label_entries


def has_overlapping_vulnerability_id(label_entry: LabelEntry, match: Match) -> bool:
    left_ids = {label_entry.vulnerability_id, label_entry.effective_cve}
    right_ids = {match.vulnerability.id, match.vulnerability.cve_id}

    if "" in left_ids:
        left_ids.remove("")

    if "" in right_ids:
        right_ids.remove("")

    return bool(left_ids & right_ids)


def _contains_as_value(o, target):
    if isinstance(o, dict):
        values = [v for k, v in o.items()]
    elif isinstance(o, list):
        values = o
    else:
        return target in o
    for v in values:
        if v == target:
            return True
        if isinstance(v, dict) and _contains_as_value(v, target):
            return True
        if isinstance(v, list) and _contains_as_value(v, target):
            return True
    return False


def merge_label_entries(
    original_entries: List[LabelEntry],
    new_and_modified_entries: List[LabelEntry],
    deleted_ids: Optional[List[str]] = None,
) -> List[LabelEntry]:
    # keep a copy to prevent mutating the argument
    original_entries = original_entries[:]
    new_and_modified_entries = new_and_modified_entries[:]

    # keep list indexes by label entry IDs
    new_and_modified_id_idx = {le.ID: idx for idx, le in enumerate(new_and_modified_entries)}

    # step 1: take potentially mutated entries and overwrite the original entries
    for original_idx, _ in enumerate(original_entries):
        original_id = original_entries[original_idx].ID
        if original_id in new_and_modified_id_idx:
            # overwrite the old entry with the new data
            new_idx = new_and_modified_id_idx[original_id]
            original_entries[original_idx] = new_and_modified_entries[new_idx]
            # mark this index as not new for later skips
            new_and_modified_entries[new_idx] = None  # type: ignore[call-overload]

    # step 2: there may be entries that are None, which should be deleted (unexpected, but possible)
    original_entries = [entry for entry in original_entries if entry]

    # step 3: everything left behind is a new entry, add it to the list
    new_entries = [entry for entry in new_and_modified_entries if entry]

    # step 4: remove all ids which are explicitly deleted
    if deleted_ids:
        original_entries = [entry for entry in original_entries if entry.ID not in deleted_ids]

    # add the new elements to the final result
    return original_entries + new_entries



from .delta import DeltaType, Delta
from .gate import Gate, GateConfig, GateInputResultConfig, GateInputDescription
from .validate import validate_image, validate_result_set

__all__ = [
    "GateConfig",
    "GateInputResultConfig",
    "GateInputDescription",
    "DeltaType",
    "Delta",
    "Gate",
    "validate_image",
    "validate_result_set",
]



from __future__ import annotations

import datetime
import logging
import subprocess
import sys
from dataclasses import dataclass

import click
from tabulate import tabulate

import yardstick
from yardstick import artifact, capture, store
from yardstick.cli import config, display, explore


@click.group(name="result", help="manage image scan results")
@click.pass_obj
def group(
    _: config.Application,
):
    pass


@group.command(name="capture", help="capture all tool output for the given image")
@click.option("--image", "-i", help="the image to scan")
@click.option("--tool", "-t", help="the tool 'name@version' reference")
@click.option(
    "--profile",
    "-p",
    help="an optional profile name of extra install/capture configuration",
)
@click.option(
    "--result-set",
    "-r",
    help="use a named result set instead of a single tool/image",
)
@click.option(
    "--only-producers",
    help="only capture results from producers",
    is_flag=True,
)
@click.pass_obj
def capture_results(  # noqa: PLR0913
    cfg: config.Application,
    image: str,
    tool: str,
    profile: str,
    result_set: str,
    only_producers: bool,
):
    if (image or tool or profile) and result_set:
        raise RuntimeError("cannot specify both image+tool and result-set, choose one")

    if not ((image and tool) or result_set):
        raise RuntimeError("must specify image and tool OR result-set")

    if result_set:
        result_set_config = cfg.result_sets.get(result_set, None)
        if not result_set_config:
            raise RuntimeError(f"no result set found for {result_set}")

        try:
            capture.result_set(
                result_set,
                result_set_config.scan_requests(),
                only_producers=only_producers,
                profiles=cfg.profiles.data,
            )
        except subprocess.CalledProcessError as e:
            logging.error(f"unable to capture result set: {e.output.decode()}")
            raise e
    else:
        scan_result = capture.one(
            artifact.ScanRequest(image=image, tool=tool, profile=profile),
        )
        print(scan_result.ID)


@group.command(name="compare", help="show a comparison between tool output")
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option(
    "--summary",
    "-s",
    default=False,
    is_flag=True,
    help="show summary without detailed breakdown of each entry",
)
@click.option(
    "--show-common",
    "-c",
    default=False,
    is_flag=True,
    help="show common match details",
)
@click.argument("ids", nargs=-1)
@click.pass_obj
def compare_results(
    cfg: config.Application,
    ids: list[str],
    year_max_limit: int | None,
    summary: bool,
    show_common: bool,
):
    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    comp = yardstick.compare_results(descriptions=ids, year_max_limit=year_max_limit)

    display.preserved_matches(
        comp,
        details=not summary,
        summary=summary,
        common=show_common,
    )


@group.command(name="show", help="show a the results for a single scan + tool")
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option(
    "--derive-year-from-cve-only",
    "-c",
    default=None,
    help="only use the CVE ID year-max-limit",
    is_flag=True,
)
@click.argument("description")
@click.pass_obj
def show_results(
    cfg: config.Application,
    description: str,
    year_max_limit: int | None,
    derive_year_from_cve_only: None | bool,
):
    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    if derive_year_from_cve_only is None:
        derive_year_from_cve_only = cfg.derive_year_from_cve_only

    logging.info(
        f"showing capture data for {description} (year limit: {year_max_limit})",
    )

    scan_config = store.scan_result.find_one(by_description=description)

    result = store.scan_result.load(
        config=scan_config,
        year_max_limit=year_max_limit,
        year_from_cve_only=derive_year_from_cve_only,
    )

    if result.matches:
        for match in sorted(result.matches):
            print(match)


@group.command(name="clear", help="remove all results and result sets")
@click.pass_obj
def clear_results(_: config.Application):
    logging.info("deleting all results")

    store.scan_result.clear()
    store.result_set.clear()


@group.command(name="list", help="list stored results")
@click.option(
    "--result-set",
    "-r",
    help="filter list to results for a named result set (instead of all results)",
)
@click.option(
    "--tool",
    "-t",
    "tools",
    help="filter results down to that partially match given tool names",
    multiple=True,
)
@click.option(
    "--images",
    "-i",
    "images",
    help="filter results down to that partially match given image names",
    multiple=True,
)
@click.option("--ids", "show_id", help="show result IDs only", is_flag=True)
@click.pass_obj
def list_results(
    _: config.Application,
    result_set: str,
    show_id: bool,
    tools: list[str],
    images: list[str],
):
    results = result_descriptions(result_set=result_set)

    if tools:
        results = [r for r in results if any(t in r.tool for t in tools)]

    if images:
        results = [r for r in results if any(i in r.image for i in images)]

    if show_id:
        for result in results:
            print(result.ID)
    else:
        all_rows = []
        for result in results:
            all_rows.append([result.ID, result.image, result.tool, result.timestamp])

        print(tabulate(all_rows, tablefmt="plain"))


@group.command(name="images", help="list images in results")
@click.option(
    "--result-set",
    "-r",
    help="filter list to results for a named result set (instead of all results)",
)
@click.pass_obj
def list_images(cfg: config.Application, result_set: str):
    if result_set:
        result_set_config = cfg.result_sets.get(result_set, None)
        if not result_set_config:
            raise RuntimeError(f"no result set found for {result_set}")

    results = result_descriptions(result_set=result_set)
    images = set()
    for result in results:
        images.add(result.image)

    for image in sorted(images):
        print(image)


@group.command(name="tools", help="list tools in results")
@click.option(
    "--result-set",
    "-r",
    help="filter list to results for a named result set (instead of all results)",
)
@click.pass_obj
def list_tools(cfg: config.Application, result_set: str):
    if result_set:
        result_set_config = cfg.result_sets.get(result_set, None)
        if not result_set_config:
            raise RuntimeError(f"no result set found for {result_set}")

    results = result_descriptions(result_set=result_set)
    tools = set()
    for result in results:
        tools.add(result.tool)

    for tool in sorted(tools):
        print(tool)


@dataclass(eq=True, order=True)
class ResultDescription:
    image: str
    tool: str
    ID: str
    timestamp: datetime.datetime | None = None


def result_descriptions(result_set: str | None = None) -> list[ResultDescription]:
    results = []
    if result_set:
        result_set_obj = store.result_set.load(result_set)

        for result_state in result_set_obj.state:
            if not result_state.config:
                raise ValueError("result set missing a configuration")

            scan_config = store.scan_result.find_one(
                by_description=result_state.config.path,
            )

            if not scan_config:
                raise RuntimeError(
                    f"unable to find scan configuration for {result_state.config.path}",
                )

            results.append(
                ResultDescription(
                    ID=scan_config.ID,
                    image=result_state.request.image,
                    tool=result_state.config.tool,
                    timestamp=result_state.config.timestamp,
                ),
            )

    else:
        scan_configs = store.scan_result.list_all_configs()
        for scan_config in scan_configs:
            results.append(
                ResultDescription(
                    ID=scan_config.ID,
                    image=scan_config.image,
                    tool=scan_config.tool,
                    timestamp=scan_config.timestamp,
                ),
            )

    return sorted(results)


@group.group(name="set", help="manipulate result sets")
@click.pass_obj
def set_group(_: config.Application):
    pass


@set_group.command(name="list", help="list configured result sets")
@click.pass_obj
def list_result_sets(_: config.Application):
    for result_set in store.result_set.load_all():
        print(result_set.name)


@set_group.command(name="add", help="create a result set")
@click.argument("ids", nargs=-1)
@click.option(
    "--name",
    "-n",
    "result_set",
    required=True,
    help="the name of the result set",
)
@click.pass_obj
def add_result_sets(_: config.Application, ids: list[str], result_set: str):
    result_set_obj = artifact.ResultSet(name=result_set)
    for scan_config_id in ids:
        scan_config = store.scan_result.find_one(by_description=scan_config_id)
        scan_request = artifact.ScanRequest(
            image=scan_config.full_image,
            tool=scan_config.tool,
        )

        if not scan_config:
            raise RuntimeError(f"unable to find scan configuration for {scan_request}")

        result_set_obj.add(request=scan_request, scan_config=scan_config)
    store.result_set.save(result_set_obj)


@group.command(name="import", help="import results for a tool that were run externally")
@click.option("--image", "-i", required=True, help="the image that was scanned")
@click.option("--tool", "-t", required=True, help="the tool 'name@version' reference")
@click.option(
    "--file",
    "-f",
    required=False,
    help="the file path to the image scan results (uses stdin if not provided)",
)
@click.pass_obj
def import_results(
    _: config.Application,
    image: str,
    tool: str,
    file: str | None = None,
):
    logging.info(f"importing data image={image} tool={tool} file={file}")

    scan_config = artifact.ScanConfiguration.new(
        image=image,
        tool=tool,
        timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
    )

    if file:
        with open(file, encoding="utf-8") as file_handle:
            raw_results = file_handle.read()
    else:
        logging.info("reading stdin")
        raw_results = sys.stdin.read()

    match_results = capture.intake(config=scan_config, raw_results=raw_results)
    store.scan_result.save(raw_results, match_results)
    print(scan_config.ID)


@group.command(name="explore", help="interact with an image scan result")
@click.argument("ids")
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option(
    "--derive-year-from-cve-only",
    "-c",
    default=None,
    help="only use the CVE ID year-max-limit",
    is_flag=True,
)
@click.pass_obj
def explore_results(
    cfg: config.Application,
    ids: str,
    year_max_limit: int | None,
    derive_year_from_cve_only: None | bool,
):
    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    if derive_year_from_cve_only is None:
        derive_year_from_cve_only = cfg.derive_year_from_cve_only

    scan_config = store.scan_result.find_one(by_description=ids)
    result = store.scan_result.load(config=scan_config)
    if year_max_limit:
        results = store.scan_result.filter_by_year(
            [result],
            int(year_max_limit),
            year_from_cve_only=derive_year_from_cve_only,
        )
        result = results[0]
    explore.result.run(result)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from yardstick.store import config as store_config

if TYPE_CHECKING:
    from yardstick import artifact

TOOL_DIR = "tools"
RESULT_DIR = os.path.join("result", "store")
RESULT_SET_DIR = os.path.join("result", "sets")


def install_base(name: str, store_root: str | None = None) -> str:
    if not store_root:
        store_root = store_config.get().store_root

    return os.path.join(store_config.get().store_root, TOOL_DIR, name.replace("/", "_"))


def install_path(
    config: artifact.ScanConfiguration,
    store_root: str | None = None,
) -> str:
    if not store_root:
        store_root = store_config.get().store_root

    store_base = install_base(name=config.tool_name, store_root=store_root)

    version = config.tool_version
    if config.tool_name.startswith("grype"):
        version = version.split("+import-db=")[0]

    return os.path.join(
        store_base,
        version.replace("/", "_"),
    )


def results_path(store_root: str | None = None):
    if not store_root:
        store_root = store_config.get().store_root

    return os.path.join(store_root, RESULT_DIR)


def result_set_path(store_root: str | None = None):
    if not store_root:
        store_root = store_config.get().store_root

    return os.path.join(store_root, RESULT_SET_DIR)



from yardstick import utils


class CveDescriptions:
    def __init__(self):
        self.grype_db = utils.grype_db.GrypeDBManager()
        self.cache = {}

    def is_cached(self, cve: str):
        return cve in self.cache

    def get(self, cve: str):
        if cve in self.cache:
            return self.cache[cve]

        if not self.grype_db.enabled:
            description = "could not connect to grype db:\n" + self.grype_db.message
        else:
            description = self.grype_db.get_all_vulnerability_descriptions(cve)

        if not description:
            description = "Unable to get CVE description"

        self.cache[cve] = description
        return description



from __future__ import annotations

import hashlib
import logging
import os

import git


def local_build_version_suffix(src_path: str) -> str:
    src_path = os.path.abspath(os.path.expanduser(src_path))
    git_desc = ""
    diff_digest = "clean"
    try:
        repo = git.Repo(src_path)
    except:
        logging.error(f"failed to open existing repo at {src_path!r}")
        raise
    git_desc = repo.git.describe("--tags", "--always", "--long", "--dirty")
    if repo.is_dirty():
        # note on S324 usage: this is currently only used for deriving a unique, content-sensitive
        # value to use for identifying local builds. This is not used for cryptographic purposes.
        hash_obj = hashlib.sha1()  # noqa: S324
        for untracked in repo.untracked_files:
            hash_obj.update(
                hash_file(os.path.join(repo.working_dir, untracked)).encode(),
            )
        hash_obj.update(repo.git.diff("HEAD").encode())
        diff_digest = hash_obj.hexdigest()[:8]
    return f"{git_desc}-{diff_digest}"


def hash_file(path: str) -> str:
    # note on S324 usage: this is currently only used for deriving a unique, content-sensitive
    # value to use for identifying local builds. This is not used for cryptographic purposes.
    hash_obj = hashlib.sha1()  # noqa: S324
    with open(path, "rb") as f:
        while True:
            data = f.read(4096)
            if not data:
                break
            hash_obj.update(data)
    return hash_obj.hexdigest()


def dig(target, *keys, **kwargs):
    """
    Traverse a nested set of dictionaries, tuples, or lists similar to ruby's dig function.
    """
    end_of_chain = target
    for key in keys:
        if (isinstance(end_of_chain, dict) and key in end_of_chain) or (isinstance(end_of_chain, (list, tuple)) and isinstance(key, int)):
            end_of_chain = end_of_chain[key]
        else:
            if "fail" in kwargs and kwargs["fail"] is True:
                if isinstance(end_of_chain, dict):
                    raise KeyError
                raise IndexError
            if "default" in kwargs:
                return kwargs["default"]
            end_of_chain = None
            break

    # we may have found a falsy value in the collection at the given key
    # and the caller has specified to return a default value in this case in it's place.
    if not end_of_chain and "falsy_default" in kwargs:
        end_of_chain = kwargs["falsy_default"]

    return end_of_chain


def safe_div(one, two):
    if two == 0:
        return 0
    return float(one) / float(two)


# CVE prefix + Year + Arbitrary Digits
# CVE-YYYY-NNNNN
def is_cve_vuln_id(vuln_id: str | None) -> bool:
    if not vuln_id:
        return False
    return vuln_id.lower().startswith("cve-")


def parse_year_from_id(vuln_id: str) -> int | None:
    def try_convert_year(s: str) -> int | None:
        try:
            value = int(s)
            if value < 1990 or digits_in_number(value) != 4:
                return None
            return value
        except ValueError:
            return None

    components = vuln_id.split("-")
    if not components:
        return None

    first_component = components[0].lower()

    if len(components) == 3 and first_component in {"cve", "alas", "elsa"}:
        return try_convert_year(components[1])

    # there are cases in the amazon data that are considered "extras" and the vulnerability ID is augmented
    # in a way that portrays the application scope. For instance, ALASRUBY3.0-2023-003 or ALASSELINUX-NG-2023-001.
    # fore more information on the "extras" feature for amazon linux, see: https://aws.amazon.com/amazon-linux-2/faqs/#Amazon_Linux_Extras
    if first_component.startswith("alas") and len(components) >= 3:
        # note that we need to reference the compoents from the end since the ID may contain a dynamic number of hyphens.
        return try_convert_year(components[-2])

    return None


def digits_in_number(n: int) -> int:
    count = 0
    while n > 0:
        count = count + 1
        n = n // 10
    return count



from __future__ import annotations

import logging

import click

import yardstick
from yardstick import artifact, label, store
from yardstick.cli import config, display, explore


@click.group(name="label", help="manage match labels")
@click.pass_obj
def group(_: config.Application):
    pass


@group.command(
    "compare",
    help="compare a scan result against labeled data",
)
@click.argument("descriptions", nargs=-1)
@click.option(
    "--show-fns",
    default=False,
    is_flag=True,
    help="show all FN labels for each tool-image pair",
)
@click.option(
    "--show-indeterminates",
    default=False,
    is_flag=True,
    help="show each indeterminate result for each tool-image pair",
)
@click.option(
    "--fuzzy",
    default=False,
    is_flag=True,
    help="loosen restrictions on package matching",
)
@click.option("--result-set", "-r", help="use a named result set as description input")
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option("--json", "-j", "is_json", help="show results as JSON", is_flag=True)
@click.pass_obj
def compare_results_against_labels(  # noqa: PLR0913
    cfg: config.Application,
    descriptions: list[str],
    show_fns: bool,
    show_indeterminates: bool,
    fuzzy: bool,
    result_set: str,
    year_max_limit: int | None,
    is_json: bool,
):
    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    (
        results,
        _,
        comparisons_by_result_id,
        stats_by_image_tool_pair,
    ) = yardstick.compare_results_against_labels(
        descriptions=descriptions,
        result_set=result_set,
        fuzzy=fuzzy,
        year_max_limit=year_max_limit,
    )

    if is_json:
        display.label_comparison_json(
            results,
            comparisons_by_result_id,
            stats_by_image_tool_pair,
            show_fns=show_fns,
            show_indeterminates=show_indeterminates,
        )
    else:
        display.label_comparison(
            results,
            comparisons_by_result_id,
            stats_by_image_tool_pair,
            show_fns=show_fns,
            show_indeterminates=show_indeterminates,
        )


@group.command(name="list", help="show all labels")
@click.option("--image", "-i", default=None, help="an image to filter labels on")
@click.option("--summarize", "-s", is_flag=True, help="summarize each entry")
@click.pass_obj
def list_labels(_: config.Application, image: str, summarize: bool):
    display_label_entries = store.labels.load_for_image(image) if image else store.labels.load_all()

    for entry in display_label_entries:
        if summarize:
            print(entry.summarize())
        else:
            print(entry)

    logging.info(f"total label entries: {len(display_label_entries)}")


@group.command(name="images", help="show all images derived from label data")
@click.pass_obj
def list_images(_: config.Application):
    # TODO: base this off of the label directory structure instead of the entries for better performance
    display_label_entries = store.labels.load_all()

    images = set()
    for entry in display_label_entries:
        if entry.image and entry.image.exact:
            images.add(entry.image.exact)

    for image in sorted(images):
        print(image)

    logging.info(f"total images: {len(images)}")


@group.command(name="add", help="add a match label indication for an image")
@click.option("--image", "-i", required=True, help="the image to use")
@click.option("--vulnerability", "-c", required=True, help="the vulnerability id")
@click.option("--package-name", "-p", required=True, help="the package name")
@click.option("--package-version", "-v", required=True, help="the package version")
@click.option(
    "--label",
    "-l",
    "label_name",
    required=True,
    help="the match label (tp/fp/unclear)",
)
@click.option("--note", "-n", help="an optional note")
@click.pass_obj
def add_label(  # noqa: PLR0913
    _: config.Application,
    image: str,
    vulnerability: str,
    package_name: str,
    package_version: str,
    label_name: str,
    note: str,
):
    package = artifact.Package(
        name=package_name,
        version=package_version,
    )

    label_obj = artifact.Label.from_str(label_name)
    if not label_obj:
        raise RuntimeError(f"unable to parse label={label_name!r}")

    new_label = artifact.LabelEntry(
        vulnerability_id=vulnerability,
        image=artifact.ImageSpecifier(exact=image),
        package=package,
        label=label_obj,
        note=note,
        lookup_effective_cve=True,
    )
    store.labels.save([new_label])
    print(new_label.ID)


@group.command(name="remove", help="remove a match label indication for an image")
@click.argument("label-ids", nargs=-1)
@click.pass_obj
def remove_label(
    _: config.Application,
    label_ids: list[str],
):
    deleted_ids = store.labels.delete(label_ids)
    for d in deleted_ids:
        print(d)
    logging.info(f"removed {len(deleted_ids)} labels")


@group.command(
    name="explore",
    help="interact with an label results for a single image scan",
)
@click.argument("result_id")
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option(
    "--derive-year-from-cve-only",
    "-c",
    default=None,
    help="only use the CVE ID year-max-limit",
    is_flag=True,
)
@click.pass_obj
def explore_labels(
    cfg: config.Application,
    result_id: str,
    year_max_limit: int | None,
    derive_year_from_cve_only: bool | None,
):
    logging.disable(level=logging.CRITICAL)

    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    if derive_year_from_cve_only is None:
        derive_year_from_cve_only = cfg.derive_year_from_cve_only

    scan_config = store.scan_result.find_one(by_description=result_id)
    result = store.scan_result.load(
        config=scan_config,
        year_max_limit=year_max_limit,
        year_from_cve_only=derive_year_from_cve_only,
    )

    lineage = store.image_lineage.get(scan_config.image)
    label_entries = store.labels.load_for_image(
        [scan_config.image, *lineage],
        year_max_limit=year_max_limit,
        year_from_cve_only=derive_year_from_cve_only,
    )

    filter_spec = ""
    if year_max_limit:
        filter_spec = f"CVE year <= {year_max_limit}"

    explore.image_labels.run(result, label_entries, lineage, filter_spec)


@group.command(name="set-image-parent", help="set the parent image for a given image")
@click.option("--child", "-c", help="the image that is the child")
@click.option("--parent", "-p", help="the image that is the parent")
@click.pass_obj
def set_image_parent(_: config.Application, child: str, parent: str):
    logging.info(f"setting image={child!r} parent to {parent!r}")
    store.image_lineage.add(child, [parent])
    lineage = store.image_lineage.get(child)
    logging.info(f"full image={child!r} lineage={lineage!r}")


@group.command(
    name="show-image-lineage",
    help="show all parents and children for the given image",
)
@click.option("--image", "-i", help="the container image")
@click.pass_obj
def show_image_lineage(_: config.Application, image: str):
    lineage = store.image_lineage.get(image)
    logging.info(f"full image={image!r} lineage={lineage!r}")


@group.command(
    name="apply",
    help="see which labels apply to the given image and tool pair",
)
@click.argument("result_id")
@click.option(
    "--inverse",
    "-i",
    help="show image lables that should not be applied",
    is_flag=True,
)
@click.option("--id", "show_ids", help="show IDs only", is_flag=True)
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option(
    "--derive-year-from-cve-only",
    "-c",
    default=None,
    help="only use the CVE ID year-max-limit",
    is_flag=True,
)
@click.pass_obj
def apply_labels(  # noqa: C901, PLR0913
    cfg: config.Application,
    result_id: str,
    inverse: bool,
    show_ids: bool,
    year_max_limit: int | None,
    derive_year_from_cve_only: bool | None,
):
    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    if derive_year_from_cve_only is None:
        derive_year_from_cve_only = cfg.derive_year_from_cve_only

    logging.info(f"applying labels to {result_id!r} (year limit: {year_max_limit})")

    scan_config = store.scan_result.find_one(by_description=result_id)
    result = store.scan_result.load(config=scan_config)

    lineage = store.image_lineage.get(scan_config.image)
    images = [scan_config.image, *lineage]
    label_entries_for_images = store.labels.load_for_image(
        images,
        year_max_limit=year_max_limit,
        year_from_cve_only=derive_year_from_cve_only,
    )
    label_entries = {label.ID: label for label in label_entries_for_images}

    labels = []
    for i in images:
        label_entries_for_image = store.labels.load_for_image(
            i,
            year_max_limit=year_max_limit,
            year_from_cve_only=derive_year_from_cve_only,
        )
        labels.append(label_entries_for_image)

    found = {}
    for idx, _ in enumerate(images):
        remaining_label_entries_for_image = labels[idx]

        if result.matches is None:
            raise ValueError("no matches found in result")

        for match in result.matches:
            # look through ancestors from the most root image upwards (not top down)
            paired_labels = label.find_labels_for_match(
                image=None,  # with must_match_image=False this should not matter
                match=match,
                label_entries=remaining_label_entries_for_image,
                # don't consider image field when pairing labels with matches
                must_match_image=False,
            )

            for paired_label in paired_labels:
                found[paired_label.ID] = paired_label
                if paired_label.ID in label_entries:
                    label_entries.pop(paired_label.ID)

    def show(label_entry):
        if show_ids:
            print(label_entry.ID)
        else:
            print(label_entry)

    if inverse:
        for label_entry in label_entries.values():
            show(label_entry)
    else:
        for _, label_entry in found.items():
            show(label_entry)

    logging.info(f"found {len(found)} labels that apply to {result_id!r}")


@group.command(
    "compare-by-ecosystem",
    help="show TPs/FPs/Precision from label comparison results broken down by ecosystem",
)
@click.option("--result-set", "-r", help="use a named result set as description input")
@click.option(
    "--year-max-limit",
    "-y",
    default=None,
    help="max year to include in comparison (relative to the CVE ID)",
)
@click.option(
    "--derive-year-from-cve-only",
    "-c",
    default=None,
    help="only use the CVE ID year-max-limit",
    is_flag=True,
)
@click.pass_obj
def compare_results_against_labels_by_ecosystem(
    cfg: config.Application,
    result_set: str,
    year_max_limit: int | None,
    derive_year_from_cve_only: bool | None,
):
    if not year_max_limit:
        year_max_limit = cfg.default_max_year

    if derive_year_from_cve_only is None:
        derive_year_from_cve_only = cfg.derive_year_from_cve_only

    results_by_image, _, stats = yardstick.compare_results_against_labels_by_ecosystem(
        result_set=result_set,
        year_max_limit=year_max_limit,
        year_from_cve_only=derive_year_from_cve_only,
    )

    display.labels_by_ecosystem_comparison(
        results_by_image,
        stats,
    )



import logging
import sys
from typing import Sequence, Optional, Callable

import yardstick
from yardstick import artifact, store, utils
from yardstick.cli import display
from yardstick.validate.delta import compute_deltas
from yardstick.validate.gate import (
    GateInputDescription,
    GateInputResultConfig,
    GateConfig,
    Gate,
)


def guess_tool_orientation(tools: list[str]):
    """
    Given a pair of tools, guess which is latest version, and which is the one
    being compared to the latest version. This should only be used as a fallback.
    Instead, specify reference tool label and candidate tool label in validations.
    Returns (latest_tool, current_tool)
    """
    if len(tools) != 2:
        raise RuntimeError("expected 2 tools, got %s" % tools)
    tool_a, tool_b = sorted(tools)
    if tool_a == tool_b:
        raise ValueError("latest release tool and current tool are the same")
    if tool_a.endswith("latest"):
        return tool_a, tool_b
    elif tool_b.endswith("latest"):
        return tool_b, tool_a

    if "@path:" in tool_a and "@path:" not in tool_b:
        # tool_a is a local build, so compare it against tool_b
        return tool_b, tool_a

    if "@path:" in tool_b and "@path:" not in tool_a:
        # tool_b is a local build, so compare it against tool_a
        return tool_a, tool_b

    return tool_a, tool_b


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    RESET = "\033[0m"


if not sys.stdout.isatty():
    bcolors.HEADER = ""
    bcolors.OKBLUE = ""
    bcolors.OKCYAN = ""
    bcolors.OKGREEN = ""
    bcolors.WARNING = ""
    bcolors.FAIL = ""
    bcolors.BOLD = ""
    bcolors.UNDERLINE = ""
    bcolors.RESET = ""


def results_used(image: str, results: Sequence[artifact.ScanResult]) -> GateInputDescription:
    return GateInputDescription(
        image=image,
        configs=[
            GateInputResultConfig(
                id=result.ID,
                tool=result.config.tool,
                tool_label=result.config.tool_label,
            )
            for result in results
        ],
    )


def validate_result_set(
    gate_config: GateConfig,
    result_set: str,
    images: list[str],
    always_run_label_comparison: bool,
    verbosity: int,
    label_entries: Optional[list[artifact.LabelEntry]] = None,
) -> list[Gate]:
    result_set_obj = store.result_set.load(name=result_set)

    if gate_config.allowed_namespaces:
        m_filter = namespace_filter(gate_config.allowed_namespaces)
        logging.info(f"only considering matches from allowed namespaces: {' '.join(gate_config.allowed_namespaces)}")
    else:
        m_filter = None

    ret = []
    for image, result_states in result_set_obj.result_state_by_image.items():
        if images and image not in images:
            logging.info(f"Skipping image {image!r} because --images is passed but does not include it")
            continue
        tools = ", ".join([s.request.tool for s in result_states])
        logging.info(f"Testing image: {image!r} with {tools!r}")

        gate = validate_image(
            image=image,
            gate_config=gate_config,
            descriptions=[s.config.path for s in result_states if s.config is not None],
            always_run_label_comparison=always_run_label_comparison,
            verbosity=verbosity,
            label_entries=label_entries,
            match_filter=m_filter,
        )
        ret.append(gate)

    return ret


def namespace_filter(
    namespaces: list[str],
) -> Callable[[list[artifact.Match]], list[artifact.Match]]:
    include = set(namespaces)

    def match_filter(matches: list[artifact.Match]) -> list[artifact.Match]:
        result = []
        for match in matches:
            if utils.dig(match.fullentry, "vulnerability", "namespace") in include:
                result.append(match)
        return result

    return match_filter


def validate_image(
    image: str,
    gate_config: GateConfig,
    descriptions: list[str],
    always_run_label_comparison: bool,
    verbosity: int,
    label_entries: Optional[list[artifact.LabelEntry]] = None,
    match_filter: Callable[[list[artifact.Match]], list[artifact.Match]] | None = None,
) -> Gate:
    """
    Compare the results of two different vulnerability scanner configurations with each other,
    and if necessary with label information. Returns a pass-fail Gate based on
    the comparison, which fails if the candidate tool results are worse than the reference
    tool results, as specified by the `gate_config`.

    Parameters
    ----------
    image : str
        The identifier or name of the image being analyzed.
    gate_config : GateConfig
        The configuration object that specifies comparison thresholds, tool labels,
        and allowed/required namespaces.
    descriptions : list[str]
        A list of descriptions or metadata associated with the image results for the comparison.
    always_run_label_comparison : bool
        If True, run comparison against labels even if no differences are found between the
        two tools.
    verbosity : int
        Level of verbosity for displaying comparison details. A higher value means more detailed output.
    label_entries : Optional[list[artifact.LabelEntry]], optional
        To save time, pass label entries. If present, will be used instead of loading from disk.
    match_filter : Callable[[list[artifact.Match]], list[artifact.Match]] | None, optional
        An optional filter function to refine the set of matches used in the comparison, by default None.
        Useful for filtering by namespace, for example.

    Returns
    -------
    Gate
        A `Gate` object that represents the pass/fail status based on the comparison. If the candidate
        tool results are worse than the reference tool according to the `gate_config`, the gate will fail.
        Otherwise, the gate will pass.

    Raises
    ------
    RuntimeError
        If an unexpected number of results (other than 2) are found during the label comparison.
    """
    # Load the relative comparison between the reference and candidate tool runs, without label info.
    # This optimizes performance by allowing early exit if there are no matches or identical results.
    relative_comparison = yardstick.compare_results(
        descriptions=descriptions,
        year_max_limit=gate_config.max_year,
        matches_filter=match_filter,
    )

    # show the relative comparison results
    if verbosity > 0:
        details = verbosity > 1
        display.preserved_matches(relative_comparison, details=details, summary=True, common=False)

    if gate_config.fail_on_empty_match_set:
        if not sum(len(res.matches) if res.matches else 0 for res in relative_comparison.results):
            return Gate.failing(
                reasons=[
                    "gate configured to fail on empty matches, and no matches found",
                ],
                input_description=results_used(image, relative_comparison.results),
            )

    if not always_run_label_comparison and not sum([len(relative_comparison.unique[result.ID]) for result in relative_comparison.results]):
        return Gate.passing(
            input_description=results_used(image, relative_comparison.results),
        )

    logging.info("Running comparison against labels...")
    # Compare against labels. Because the reference tool configuration and the
    # candidate tool configuration both found matches, and did not find the same
    # set of matches, we need to compare to known-correct label data and do
    # a little stats to determine whether candidate tool is better or the same
    # as reference tool.
    results, label_entries, comparisons_by_result_id, stats_by_image_tool_pair = yardstick.compare_results_against_labels(
        descriptions=descriptions,
        year_max_limit=gate_config.max_year,
        label_entries=label_entries,
        matches_filter=match_filter,
    )

    if verbosity > 0:
        show_fns = verbosity > 1
        display.label_comparison(
            results,
            comparisons_by_result_id,
            stats_by_image_tool_pair,
            show_fns=show_fns,
            show_summaries=True,
        )

    if len(results) != 2:
        raise RuntimeError(f"validate_image compares results of exactly 2 runs, but found {len(results)}")

    candidate_tool, reference_tool = tool_designations(gate_config.candidate_tool_label, [r.config for r in results])

    # keep a list of differences between tools to summarize in UI
    # not that this is different from the statistical comparison;
    # deltas basically a UI/logging concern; the stats are a pass/fail concern.
    deltas = compute_deltas(comparisons_by_result_id, reference_tool, relative_comparison)

    reference_comparisons_by_images = {comp.config.image: comp for comp in comparisons_by_result_id.values() if comp.config.tool == reference_tool}
    reference_comparison = reference_comparisons_by_images[image]
    candidate_comparisons_by_images = {comp.config.image: comp for comp in comparisons_by_result_id.values() if comp.config.tool == candidate_tool}
    candidate_comparison = candidate_comparisons_by_images[image]
    return Gate(
        reference_comparison=reference_comparison.summary,
        candidate_comparison=candidate_comparison.summary,
        config=gate_config,
        input_description=results_used(image, relative_comparison.results),
        deltas=deltas,
    )


def tool_designations(candidate_tool_label: str, scan_configs: list[artifact.ScanConfiguration]) -> tuple[str, str]:
    reference_tool, candidate_tool = None, None
    if not candidate_tool_label:
        reference_tool, candidate_tool = guess_tool_orientation(
            [config.tool for config in scan_configs],
        )
        logging.warning(f"guessed tool orientation reference:{reference_tool} candidate:{candidate_tool}")
        logging.warning("to avoid guessing, specify reference_tool_label and candidate_tool_label in validation config and re-capture result set")
    if scan_configs[0].tool_label == candidate_tool_label:
        candidate_tool = scan_configs[0].tool
        reference_tool = scan_configs[1].tool
    elif scan_configs[1].tool_label == candidate_tool_label:
        candidate_tool = scan_configs[1].tool
        reference_tool = scan_configs[0].tool
    return candidate_tool, reference_tool



from asyncio import Future

from prompt_toolkit.application.current import get_app
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Dimension
from prompt_toolkit.layout.containers import Container, HSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.widgets import Button, Dialog, TextArea
from pygments.lexers.data import JsonLexer

from yardstick import artifact


class LabelJsonEditorDialog:
    future: Future

    def __init__(
        self,
        entry: artifact.LabelEntry,
        title="Edit label JSON",
        ok_button_text="Save",
        cancel_button_text="Cancel",
    ):
        self.future = Future()

        def accept_text(buf):
            get_app().layout.focus(ok_button)
            buf.complete_state = None
            return True

        def accept():
            if self.validation_toolbar.is_value:
                self.future.set_result(self.text_area.text)

        kb = KeyBindings()

        kb.add("escape")

        def cancel():
            self.future.set_result(None)

        formatted_json = entry.to_json(indent=2)  # type: ignore[attr-defined]

        ok_button = Button(
            text=ok_button_text,
            left_symbol="[",
            right_symbol="]",
            handler=accept,
        )
        cancel_button = Button(
            text=cancel_button_text,
            left_symbol="[",
            right_symbol="]",
            handler=cancel,
        )

        self.text_area = TextArea(
            text=formatted_json,
            multiline=True,
            line_numbers=True,
            scrollbar=True,
            lexer=PygmentsLexer(JsonLexer),
            height=Dimension(preferred=50),
            accept_handler=accept_text,
            style="bg:#333333",
            complete_while_typing=True,
        )

        self.validation_toolbar = ValidationToolbar(self.text_area)

        layout_items = [self.text_area, self.validation_toolbar]

        self.dialog = Dialog(
            title=title,
            # layout_items is a list of MagicContainer, which is only exported when checking types unfortunately
            body=HSplit(layout_items, key_bindings=kb),  # type: ignore[arg-type]
            buttons=[ok_button, cancel_button],
            width=Dimension(preferred=100),
            modal=True,
        )

    def __pt_container__(self):
        return self.dialog


class ValidationToolbar:
    def __init__(self, text_area) -> None:
        self.is_value = True

        def get_formatted_text() -> StyleAndTextTuples:
            try:
                artifact.LabelEntry.from_json(  # type: ignore[attr-defined]
                    text_area.text,
                )
                self.is_value = True
            except Exception as e:
                self.is_value = False
                return [("class:validation-toolbar", "Invalid LabelEntry: " + repr(e))]

            return []

        self.control = FormattedTextControl(get_formatted_text)

        self.container = Window(
            self.control,
            height=1,
            always_hide_cursor=True,
            wrap_lines=True,
        )

    def __pt_container__(self) -> Container:
        return self.container



from dataclasses import dataclass
from typing import Callable, List


@dataclass
class Command:
    undo: Callable
    redo: Callable


class History:
    history: List[Command]
    index: int

    def __init__(self):
        self.history = []
        self.index = 0

    def record(self, command: Command):
        # clear any history after the current index, but only if the index has moved (we've done undo's since the last record)
        if self.index + 1 != len(self.history):
            self.history = self.history[: self.index]
        # record the new event
        self.history.append(command)
        self.redo()  # perform the action for the first time

    def undo(self):
        if self.index > 0:
            self.index -= 1
            self.history[self.index].undo()

    def redo(self):
        if self.index <= len(self.history) - 1:
            self.history[self.index].redo()
            self.index += 1

    def total_events(self):
        return len(self.history)

    def undone_events(self):
        return len(self.history) - (self.index)

    def reset(self):
        self.history = []
        self.index = 0



import pytest
from yardstick.cli.explore.image_labels.history import Command, History


@pytest.fixture()
def counter_command_state():
    items = []

    def redo(count):
        def do():
            print("adding", count, "to", items)
            items.append(count)

        return do

    def undo(count):
        def do():
            print("removing", count, "from", items)
            items.remove(count)

        return do

    return undo, redo, items


@pytest.fixture()
def history_of_5(counter_command_state):
    undo, redo, state = counter_command_state
    history = History()

    for i in range(5):
        history.record(Command(undo=undo(i), redo=redo(i)))
    return history, undo, redo, state


def test_history_record(history_of_5):
    history, undo, redo, state = history_of_5
    assert state == [0, 1, 2, 3, 4]


def test_history_undo(history_of_5):
    history, undo, redo, state = history_of_5
    # first undo...
    history.undo()
    assert state == [0, 1, 2, 3]

    # second undo...
    history.undo()
    assert state == [0, 1, 2]

    # undo until empty...
    history.undo()
    history.undo()
    history.undo()
    assert state == []

    # check we don't go path the boundary
    history.undo()
    assert state == []


def test_history_redo(history_of_5):
    history, undo, redo, state = history_of_5
    # prep undo...
    history.undo()
    history.undo()
    history.undo()
    assert state == [0, 1]

    # first redo...
    history.redo()
    assert state == [0, 1, 2]

    # redo all...
    history.redo()
    history.redo()
    assert state == [0, 1, 2, 3, 4]

    # check we don't go path the boundary
    history.redo()
    assert state == [0, 1, 2, 3, 4]


def test_history_undo_redo_rewrite(history_of_5):
    history, undo, redo, state = history_of_5
    # prep undo...
    history.undo()
    history.undo()
    history.undo()
    assert state == [0, 1]

    # first redo...
    history.redo()
    assert state == [0, 1, 2]

    # record new history
    history.record(Command(undo=undo(42), redo=redo(42)))

    assert state == [0, 1, 2, 42]

    # redo does nothing...
    history.redo()
    history.redo()
    assert state == [0, 1, 2, 42]

    # undo still works
    history.undo()
    history.undo()
    assert state == [0, 1]

    # a few more cases...
    history.record(Command(undo=undo(72), redo=redo(72)))

    assert state == [0, 1, 72]

    history.undo()
    history.undo()
    history.undo()
    history.undo()
    assert state == []

    history.redo()
    history.redo()
    history.redo()
    history.redo()
    history.redo()
    history.redo()
    history.redo()

    assert state == [0, 1, 72]



"""
Collection of reusable components for building full screen applications.

All of these widgets implement the ``__pt_container__`` method, which makes
them usable in any situation where we are expecting a `prompt_toolkit`
container object.

.. warning::

    At this point, the API for these widgets is considered unstable, and can
    potentially change between minor releases (we try not too, but no
    guarantees are made yet). The public API in
    `prompt_toolkit.shortcuts.dialogs` on the other hand is considered stable.
"""

from typing import List, Optional

from prompt_toolkit.auto_suggest import AutoSuggest, DynamicAutoSuggest
from prompt_toolkit.buffer import Buffer, BufferAcceptHandler
from prompt_toolkit.completion import Completer, DynamicCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import (
    Condition,
    FilterOrBool,
    has_focus,
    is_done,
    is_true,
    to_filter,
)
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.history import History
from prompt_toolkit.key_binding import KeyBindingsBase
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.layout.containers import Container, Window
from prompt_toolkit.layout.controls import BufferControl, GetLinePrefixCallable
from prompt_toolkit.layout.dimension import AnyDimension
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.layout.margins import Margin, NumberedMargin, ScrollbarMargin
from prompt_toolkit.layout.processors import (
    AppendAutoSuggestion,
    BeforeInput,
    ConditionalProcessor,
    PasswordProcessor,
    Processor,
)
from prompt_toolkit.lexers import DynamicLexer, Lexer
from prompt_toolkit.validation import DynamicValidator, Validator
from prompt_toolkit.widgets.toolbars import SearchToolbar

__all__ = [
    "TextArea",
]

E = KeyPressEvent


class TextArea:
    def __init__(  # noqa: PLR0913
        self,
        text: AnyFormattedText = "",
        multiline: FilterOrBool = True,
        password: FilterOrBool = False,
        lexer: Optional[Lexer] = None,
        auto_suggest: Optional[AutoSuggest] = None,
        completer: Optional[Completer] = None,
        complete_while_typing: FilterOrBool = True,
        validator: Optional[Validator] = None,
        accept_handler: Optional[BufferAcceptHandler] = None,
        history: Optional[History] = None,
        focusable: FilterOrBool = True,
        focus_on_click: FilterOrBool = False,
        wrap_lines: FilterOrBool = True,
        read_only: FilterOrBool = False,
        width: AnyDimension = None,
        height: AnyDimension = None,
        dont_extend_height: FilterOrBool = False,
        dont_extend_width: FilterOrBool = False,
        line_numbers: bool = False,
        get_line_prefix: Optional[GetLinePrefixCallable] = None,
        scrollbar: bool = False,
        style: str = "",
        search_field: Optional[SearchToolbar] = None,
        preview_search: FilterOrBool = True,
        prompt: AnyFormattedText = "",
        input_processors: Optional[List[Processor]] = None,
        cursorline: bool = False,
        cursorcolumn: bool = False,
        key_bindings: Optional[KeyBindingsBase] = None,
        right_margins: Optional[List[Margin]] = None,
    ) -> None:
        if search_field is None:
            search_control = None
        elif isinstance(search_field, SearchToolbar):
            search_control = search_field.control

        if input_processors is None:
            input_processors = []

        # Writeable attributes.
        self.completer = completer
        self.complete_while_typing = complete_while_typing
        self.lexer = lexer
        self.auto_suggest = auto_suggest
        self.read_only = read_only
        self.wrap_lines = wrap_lines
        self.validator = validator

        self.buffer = Buffer(
            document=Document(text, 0),  # type: ignore[arg-type]
            multiline=multiline,
            read_only=Condition(lambda: is_true(self.read_only)),
            completer=DynamicCompleter(lambda: self.completer),
            complete_while_typing=Condition(
                lambda: is_true(self.complete_while_typing),
            ),
            validator=DynamicValidator(lambda: self.validator),
            auto_suggest=DynamicAutoSuggest(lambda: self.auto_suggest),
            accept_handler=accept_handler,
            history=history,
        )

        self.control = BufferControl(
            buffer=self.buffer,
            lexer=DynamicLexer(lambda: self.lexer),
            input_processors=[
                ConditionalProcessor(
                    AppendAutoSuggestion(),
                    has_focus(self.buffer) & ~is_done,
                ),
                ConditionalProcessor(
                    processor=PasswordProcessor(),
                    filter=to_filter(password),
                ),
                BeforeInput(prompt, style="class:text-area.prompt"),
                *input_processors,
            ],
            search_buffer_control=search_control,
            preview_search=preview_search,
            focusable=focusable,
            focus_on_click=focus_on_click,
            key_bindings=key_bindings,
        )

        if multiline:
            _right_margins: list[Margin] = [ScrollbarMargin(display_arrows=False)] if scrollbar else []

            if right_margins:
                _right_margins = right_margins + _right_margins

            left_margins = [NumberedMargin()] if line_numbers else []
        else:
            height = D.exact(1)
            left_margins = []
            _right_margins = []

        style = "class:text-area " + style

        # If no height was given, guarantee height of at least 1.
        if height is None:
            height = D(min=1)

        self.window = Window(
            height=height,
            width=width,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width,
            content=self.control,
            style=style,
            wrap_lines=Condition(lambda: is_true(self.wrap_lines)),
            left_margins=left_margins,
            right_margins=_right_margins,
            get_line_prefix=get_line_prefix,
            cursorline=cursorline,
            cursorcolumn=cursorcolumn,
        )

    @property
    def text(self) -> str:
        """
        The `Buffer` text.
        """
        return self.buffer.text

    @text.setter
    def text(self, value: str) -> None:
        self.document = Document(value, 0)

    @property
    def document(self) -> Document:
        """
        The `Buffer` document (text + cursor position).
        """
        return self.buffer.document

    @document.setter
    def document(self, value: Document) -> None:
        self.buffer.set_document(value, bypass_readonly=True)

    @property
    def accept_handler(self) -> Optional[BufferAcceptHandler]:
        """
        The accept handler. Called when the user accepts the input.
        """
        return self.buffer.accept_handler

    @accept_handler.setter
    def accept_handler(self, value: BufferAcceptHandler) -> None:
        self.buffer.accept_handler = value

    def __pt_container__(self) -> Container:
        return self.window



from typing import TYPE_CHECKING, Callable

from prompt_toolkit.formatted_text import StyleAndTextTuples, fragment_list_width
from prompt_toolkit.layout.controls import UIContent
from prompt_toolkit.layout.margins import Margin

if TYPE_CHECKING:
    from prompt_toolkit.layout.containers import WindowRenderInfo


class LabelMargin(Margin):
    def __init__(self, label_manager) -> None:
        self.label_manager = label_manager
        self.minwidth = 8
        self.maxwidth = 8 * 3  # no more than about 3 tags worth
        self.width = self.minwidth

    def get_width(self, get_ui_content: Callable[[], UIContent]) -> int:
        return min(self.maxwidth, self.width * 2)

    def create_margin(
        self,
        window_render_info: "WindowRenderInfo",
        width: int,
        height: int,
    ) -> StyleAndTextTuples:
        # Get current line number.
        current_lineno = window_render_info.ui_content.cursor_position.y

        # Construct margin.
        result: StyleAndTextTuples = []
        last_lineno = None

        entries = self.label_manager.match_select_entries

        self.width = self.minwidth
        for _, lineno in enumerate(window_render_info.displayed_lines):
            # Only display line number if this line is not a continuation of the previous line.
            if lineno != last_lineno:
                if lineno is None:
                    pass

                offset = 0
                if lineno == current_lineno:
                    offset = 2

                if lineno < len(entries):
                    text = entries[lineno].get_formatted_annotations()

                    content_width = fragment_list_width(text)
                    self.width = max(self.width, content_width, self.minwidth)
                    padding = width - content_width
                    if padding > 0:
                        result.append(("", " " * (padding - offset)))

                    if lineno == current_lineno:
                        result.append(("fg:#888888", "● "))
                    result.extend(text)

            last_lineno = lineno
            result.append(("", "\n"))

        return result



import json
import logging
import queue
from collections import defaultdict
from threading import Lock, Thread
from typing import Dict, List, Optional, Tuple

import pygments
from prompt_toolkit.application import get_app
from prompt_toolkit.formatted_text import (
    HTML,
    AnyFormattedText,
    FormattedText,
    PygmentsTokens,
    merge_formatted_text,
    to_formatted_text,
)
from pygments.lexers.data import JsonLexer
from tabulate import tabulate

from yardstick import artifact, comparison, store
from yardstick.cli.explore.image_labels.history import Command, History
from yardstick.cli.explore.result import MatchCollection
from yardstick.label import find_labels_for_match
from yardstick.tool import get_tool

# a set of unique locks for each object
_locks: Dict[int, Lock] = defaultdict(Lock)


def synchronized(func):
    def _synchronized(self, *args, **kw):
        with _locks[self]:
            return func(self, *args, **kw)

    return _synchronized


class WorkerWithTaskInvalidation:
    def __init__(self):
        self.done = False
        self.work: queue.Queue = queue.Queue(maxsize=0)
        self.t = Thread(target=self._worker)
        self.t.daemon = True
        self.t.start()

    def _worker(self):
        while not self.done:
            fn = self.work.get()
            try:
                fn()
            except Exception:
                logging.exception("background worker failed to complete a task")

    def submit(self, fn):
        self.clear()  # keep only the most recent request for work, invalidate anything queued
        self.work.put_nowait(fn)

    def clear(self):
        # a bit of a hack, but essentially removes all items from the queue in a thread-safe fashion
        with self.work.mutex:
            self.work.queue.clear()
            self.work.all_tasks_done.notify_all()
            self.work.unfinished_tasks = 0


class MatchSelectEntry:
    def __init__(
        self,
        match: artifact.Match,
        display: str,
        labels: Optional[AnyFormattedText] = None,
    ):
        self.match = match
        self.display = display
        self.formatted_labels = labels or FormattedText()
        self.num_labels = 0

    def get_formatted_annotations(self):
        return self.formatted_labels

    def get_formatted_details(self):
        table = [
            ["<b>Vulnerability:</b>", self.match.vulnerability.id],
            ["<b>Package Name:</b>", self.match.package.name],
            ["<b>Package Version:</b>", self.match.package.version],
            ["<b>Match ID:</b>", self.match.ID],
            ["<b>Full JSON Entry:</b>", ""],
        ]
        rows = []

        for row in tabulate(table, tablefmt="plain").split("\n"):
            rows.append(HTML(row + "\n"))

        json_str = json.dumps(self.match.fullentry, indent=2)
        tokens = list(pygments.lex(json_str, lexer=JsonLexer()))
        formatted_full_entry = to_formatted_text(PygmentsTokens(tokens))
        rows.append(formatted_full_entry)

        return merge_formatted_text(rows)


class LabelManager:
    def __init__(
        self,
        result: artifact.ScanResult,
        label_entries: List[artifact.LabelEntry],
        lineage: List[str],
    ):
        self.result = result
        self.lineage = lineage
        self.collection = MatchCollection(result)
        self.filter_text = None
        # we should filter down the set of label entries that we have to only those which match with this image. This
        # is a performance enhancement to prevent matching an ALL labels for all images on every UI action.
        self.label_entries: List[artifact.LabelEntry] = self._keep_only_matched_label_entries(label_entries, lineage)
        self.match_select_entries_invalidated = True
        self._last_match_select_entries: List[MatchSelectEntry] = []
        self.deleted_label_entries: List[artifact.LabelEntry] = []
        self.worker = WorkerWithTaskInvalidation()
        self.comparison: Optional[comparison.AgainstLabels] = None
        self.history = History()
        self.labels_invalidated = False
        self._update_comparison()

    def apply_filter(self, text):
        self.match_select_entries_invalidated = True
        self.filter_text = text

    def _keep_only_matched_label_entries(
        self,
        label_entries,
        lineage,
    ) -> List[artifact.LabelEntry]:
        keep_label_entries = []
        all_matches = self.collection.get_matches(filter_text=None)
        for match in all_matches:
            for label_entry in find_labels_for_match(
                image=self.result.config.image,
                match=match,
                label_entries=label_entries,
                lineage=lineage,
            ):
                if label_entry not in keep_label_entries:
                    keep_label_entries.append(label_entry)
        return keep_label_entries

    @property
    def match_select_entries(self) -> List[MatchSelectEntry]:
        if self.match_select_entries_invalidated:
            self.labels_invalidated = True
            self.match_select_entries_invalidated = False
            all_matches = self.collection.get_matches(filter_text=self.filter_text)

            table: List[List[str]] = []
            entries: List[MatchSelectEntry] = []

            for match in all_matches:
                if not match.config:
                    continue

                t = get_tool(match.config.tool_name)
                package_type = t.parse_package_type(match.fullentry)  # type: ignore[union-attr]
                row_cells = [
                    match.vulnerability.id,
                    f"{match.package.name} @ {match.package.version}",
                ]
                if package_type and package_type != "unknown":
                    row_cells.append(package_type)
                table.append(row_cells)

            rows = tabulate(table, tablefmt="plain").split("\n")

            for idx, match in enumerate(all_matches):
                row = rows[idx]
                select_entry = MatchSelectEntry(match=match, display=row)
                entries.append(select_entry)

            self._last_match_select_entries = entries
            self._update()

        return self._last_match_select_entries

    def _update_comparison(self):
        try:
            self.comparison = comparison.AgainstLabels(
                result=self.result,
                label_entries=self.label_entries,
                lineage=self.lineage,
            )

            # update formatted labels
            for entry in self.match_select_entries:
                entry.formatted_labels, entry.num_labels = format_labels(
                    entry.match,
                    self.comparison,
                )
        except:  # noqa: E722
            logging.exception("could not update label pane")

        get_app().invalidate()

    def _update(self):
        if self.labels_invalidated:
            self.labels_invalidated = False
            self.worker.submit(self._update_comparison)

    def get_label_entries_by_match(
        self,
        match: artifact.Match,
    ) -> List[artifact.LabelEntry]:
        return find_labels_for_match(
            image=self.result.config.image,
            match=match,
            label_entries=self.label_entries,
            lineage=self.lineage,
        )

    def get_label_entry_by_id(
        self,
        label_entry_id: str,
    ) -> Optional[artifact.LabelEntry]:
        label_entries = [label for label in self.label_entries if label_entry_id == label.ID]
        if len(label_entries) > 1:
            raise RuntimeError

        if not label_entries:
            return None

        return label_entries[0]

    def get_label_entry_index(self, label_entry_id: str) -> Optional[int]:
        for idx, label in enumerate(self.label_entries):
            if label_entry_id == label.ID:
                return idx
        return None

    @synchronized
    def edit_label_entry_json(self, label_entry_id: str, raw_json: str = ""):
        old_label_entry = self.get_label_entry_by_id(label_entry_id)

        def undo():
            entry_idx = self.get_label_entry_index(label_entry_id)
            if entry_idx is not None:
                self.label_entries[entry_idx] = old_label_entry

        def redo():
            entry_idx = self.get_label_entry_index(label_entry_id)
            if entry_idx is not None:
                self.label_entries[entry_idx] = artifact.LabelEntry.from_json(
                    raw_json,
                )

        self.history.record(Command(undo=undo, redo=redo))

        self._update()

    @synchronized
    def edit_label_entry_note(self, label_entry_id: str, note: str = ""):
        label_entry = self.get_label_entry_by_id(label_entry_id)
        if not label_entry:
            raise ValueError(f"no label entry found with ID {label_entry_id}")
        old_note = label_entry.note

        def undo():
            self.labels_invalidated = True
            label_entry.note = old_note

        def redo():
            self.labels_invalidated = True
            label_entry.note = note

        self.history.record(Command(undo=undo, redo=redo))

        self._update()

    @synchronized
    def add_label_entry(
        self,
        match: artifact.Match,
        label: artifact.Label,
        note: str = "",
    ):
        new_entry = artifact.LabelEntry(
            # source=MANUAL_SOURCE,
            image=artifact.ImageSpecifier(exact=self.result.config.image),
            vulnerability_id=match.vulnerability.id,
            package=match.package,
            label=label,
            note=note,
            tool=self.result.config.tool,
            lookup_effective_cve=True,
        )

        def undo():
            self.labels_invalidated = True
            # note: cannot use list.remove() since hash() of LabelEntry does not include ID
            self.label_entries = [e for e in self.label_entries if e.ID != new_entry.ID]

        def redo():
            self.labels_invalidated = True
            self.label_entries.append(new_entry)

        self.history.record(Command(undo=undo, redo=redo))

        self._update()

    @synchronized
    def remove_label_entry(self, label_entry_id: str):
        label_entry = self.get_label_entry_by_id(label_entry_id)
        if not label_entry:
            raise ValueError(f"no label entry found with ID {label_entry_id}")
        label_entry_index = self.label_entries.index(label_entry)

        def undo():
            self.labels_invalidated = True
            self.label_entries.insert(label_entry_index, label_entry)
            self.deleted_label_entries.remove(label_entry)

        def redo():
            self.labels_invalidated = True
            self.label_entries.remove(label_entry)
            self.deleted_label_entries.append(label_entry)

        self.history.record(Command(undo=undo, redo=redo))

        self._update()

    def total_labels(self):
        return len(self.label_entries)

    def applied_labels(self):
        applied = 0
        for entry in self.match_select_entries:
            applied += entry.num_labels
        return applied

    def matches_not_labeled(self):
        no_labels = 0
        for entry in self.match_select_entries:
            if not entry.num_labels:
                no_labels += 1
        return no_labels

    def matches_labeled(self):
        num_labeled = 0
        for entry in self.match_select_entries:
            if entry.num_labels:
                num_labeled += 1
        return num_labeled

    def f1_score(self):
        if self.comparison is not None:
            low = f"{self.comparison.summary.f1_score_lower_confidence:0.2f}"
            high = f"{self.comparison.summary.f1_score_upper_confidence:0.2f}"
            return f"{self.comparison.summary.f1_score:.2f} (possible range {low} - {high})"
        return "??"

    @synchronized
    def undo(self):
        self.history.undo()
        self._update()

    @synchronized
    def redo(self):
        self.history.redo()
        self._update()

    @synchronized
    def write(self):
        store.labels.append_and_update(
            new_and_modified_entries=self.label_entries,
            delete_entries=self.deleted_label_entries,
        )
        self.history.reset()


def format_labels(
    match: artifact.Match,
    comp: comparison.AgainstLabels,
) -> Tuple[AnyFormattedText, int]:
    labels = comp.labels_by_match[match.ID]
    colors = {
        artifact.Label.TruePositive: ("#0062ff", "#002a6e", "#ffffff"),
        artifact.Label.FalsePositive: ("#ff0066", "#99003d", "#ffffff"),
        artifact.Label.Unclear: ("#888888", "#444444", "#ffffff"),
    }

    result = []
    applied_labels = 0
    for label in (
        artifact.Label.TruePositive,
        artifact.Label.FalsePositive,
        artifact.Label.Unclear,
    ):
        c = labels.count(label)
        if c:
            applied_labels += c
            bg, dbg, fg = colors[label]
            result += [
                to_formatted_text(
                    " " + label.display + " ",
                    style=f"bold bg:{bg} {fg}",
                ),
                to_formatted_text(" " + str(c) + " ", style=f"bold bg:{dbg} {fg}"),
                to_formatted_text(" "),
            ]

    return merge_formatted_text(result)(), applied_labels  # type: ignore[operator, misc]



import abc
from typing import Any, Dict, List, Optional

from yardstick import artifact


class VulnerabilityScanner(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def install(
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ) -> "VulnerabilityScanner":
        raise NotImplementedError

    @abc.abstractmethod
    def capture(self, image: str, tool_input: Optional[str]) -> str:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def parse(result: str, config: artifact.ScanConfiguration) -> List[artifact.Match]:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def parse_package_type(full_entry: Optional[Dict[str, Any]]) -> str:
        raise NotImplementedError



import datetime
import logging
from typing import Any, Dict, Optional, Tuple, Union

from yardstick import artifact, store
from yardstick.tool import get_tool, sbom_generator, vulnerability_scanner


class Timer:
    start = None
    end = None

    def __enter__(self):
        self.start = datetime.datetime.now(datetime.timezone.utc)
        return self

    def __exit__(self, ty, value, traceback):
        self.end = datetime.datetime.now(datetime.timezone.utc)


def run_scan(
    config: artifact.ScanConfiguration,
    tool: Optional[Union[vulnerability_scanner.VulnerabilityScanner, sbom_generator.SBOMGenerator]] = None,
    reinstall: bool = False,
    **kwargs,
) -> Tuple[artifact.ScanResult, str]:
    logging.debug(
        f"capturing via run config image={config.image} tool={config.tool_name}@{config.tool_version}",
    )

    tool_cls = get_tool(str(config.tool_name))
    if not tool_cls:
        raise RuntimeError(f"unknown tool: {config.tool.name}")

    if not tool:
        path = store.tool.install_path(config=config)
        tool = tool_cls.install(
            version=config.tool_version,
            path=path,
            use_cache=not reinstall,
            **kwargs,
        )

    # some tools will have additional metadata... persist this on the config
    if hasattr(tool, "version_detail"):
        installed_version = tool.version_detail
        if installed_version != config.tool_version:
            config.detail["version_detail"] = installed_version
            config.tool_version = installed_version

    with Timer() as timer:
        raw_json = tool.capture(image=config.full_image, tool_input=config.tool_input)
        result = tool.parse(raw_json, config=config)

    config.timestamp = timer.start

    keys = {}
    if issubclass(tool_cls, vulnerability_scanner.VulnerabilityScanner):
        keys["matches"] = result
    elif issubclass(tool_cls, sbom_generator.SBOMGenerator):
        keys["packages"] = result
    else:
        raise RuntimeError("unknown tool type")

    metadata = artifact.ScanMetadata(
        timestamp=config.timestamp,
        elapsed=(timer.end - timer.start).microseconds / 100000.0,
        image_digest=config.image_digest,
    )
    return (
        artifact.ScanResult(config=config, metadata=metadata, **keys),  # type: ignore[arg-type]
        raw_json,
    )


def intake(config: artifact.ScanConfiguration, raw_results: str) -> artifact.ScanResult:
    logging.info(f"capturing via intake config={config}")

    tool_cls = get_tool(config.tool_name)
    if not tool_cls:
        raise RuntimeError(f"unknown tool: {config.tool_name}")

    result = tool_cls.parse(raw_results, config=config)
    keys = {}
    if issubclass(tool_cls, vulnerability_scanner.VulnerabilityScanner):
        keys["matches"] = result
    elif issubclass(tool_cls, sbom_generator.SBOMGenerator):
        keys["packages"] = result
    else:
        raise RuntimeError("unknown tool type")

    metadata = artifact.ScanMetadata(
        timestamp=datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0),
    )
    return artifact.ScanResult(config=config, metadata=metadata, **keys)  # type: ignore[arg-type]


def one(
    request: artifact.ScanRequest,
    producer_state: Optional[str] = None,
    profiles: Optional[Dict[str, Dict[str, Any]]] = None,
) -> artifact.ScanConfiguration:
    logging.debug(
        f"capturing data image={request.image} tool={request.tool} profile={request.profile}",
    )

    if not profiles:
        profiles = {}

    scan_config = artifact.ScanConfiguration.new(
        image=request.image,
        tool=request.tool,
        label=request.label,
    )

    if producer_state:
        scan_config.tool_input = producer_state

    profile_obj = None
    if request.profile:
        tool_name = scan_config.raw_tool_name
        profile_obj = profiles.get(tool_name, {}).get(request.profile, {})
        if not profile_obj:
            raise RuntimeError(f"no profile found for tool {tool_name}")

    match_results, raw_json = run_scan(config=scan_config, profile=profile_obj)
    store.scan_result.save(
        raw_json,
        match_results,
    )

    return scan_config


def result_set(  # noqa: C901, PLR0912
    result_set: str,
    scan_requests: list[artifact.ScanRequest],
    only_producers: bool = False,
    profiles=Optional[Dict[str, Dict[str, Any]]],
) -> artifact.ResultSet:
    logging.info(f"capturing data result_set={result_set}")

    if not profiles:
        profiles = {}

    existing_result_set_obj = None
    if store.result_set.exists(name=result_set):
        existing_result_set_obj = store.result_set.load(result_set)

    result_set_obj = artifact.ResultSet(name=result_set)
    total = len(scan_requests)
    for idx, scan_request in enumerate(scan_requests):
        logging.info(f"capturing data for request {idx + 1} of {total}")
        producer_data_path = None
        if scan_request.takes:
            producer = result_set_obj.provider(
                image=scan_request.image,
                provides=scan_request.takes,
            )
            if not producer:
                raise RuntimeError(
                    f"unable to find result state for the requested tool {scan_request}",
                )

            if producer.config:
                producer_scan_config = store.scan_result.find_one(
                    by_description=producer.config.path,
                )
                producer_data_path, _ = store.scan_result.store_paths(
                    producer_scan_config,
                )

        if only_producers and not scan_request.provides:
            logging.info(f"skipping non-producer tool {scan_request.tool}")
            continue

        refresh = scan_request.refresh
        scan_config = None

        if existing_result_set_obj and not refresh:
            result_state = existing_result_set_obj.get(
                image=scan_request.image,
                tool=scan_request.tool,
            )
            if result_state and result_state.config:
                try:
                    scan_config = store.scan_result.find_one(
                        by_description=result_state.config.path,
                    )
                except RuntimeError:
                    logging.warning(
                        f"unable to find scan config for result state, will refresh: {result_state.config.path}",
                    )
                    scan_config = None

                if scan_config:
                    logging.info(f"using existing scan result {scan_config.ID}")

        if refresh or not scan_config:
            scan_config = one(
                scan_request,
                producer_state=producer_data_path,
                profiles=profiles,
            )

        if not scan_config:
            raise RuntimeError(f"unable to find scan configuration for {scan_request}")

        result_set_obj.add(request=scan_request, scan_config=scan_config)
    store.result_set.save(result_set_obj)

    return result_set_obj



from typing import Optional, Type, Union

from .grype import Grype
from .plugin import load_plugins
from .sbom_generator import SBOMGenerator
from .syft import Syft
from .vulnerability_scanner import VulnerabilityScanner

tools: dict[str, Union[Type[SBOMGenerator], Type[VulnerabilityScanner]]] = {
    # vulnerability scanners
    "grype": Grype,
    # sbom generators
    "syft": Syft,
}


def Register(
    name: str,
    tool: Union[Type[SBOMGenerator], Type[VulnerabilityScanner]],
) -> None:
    tools[name] = tool


def get_tool(
    name: str,
) -> Optional[Union[Type[SBOMGenerator], Type[VulnerabilityScanner]]]:
    # this normalizes the name and removes labels in [brackets]
    return tools.get(name.split("[")[0].lower())


load_plugins()



from __future__ import annotations

import collections
import logging
from dataclasses import InitVar, dataclass, field
from typing import Any, Sequence

from dataclasses_json import dataclass_json
from tabulate import tabulate

from yardstick import store, utils
from yardstick.artifact import (  # MatchLabelEntry,
    Label,
    LabelEntry,
    Match,
    Package,
    ScanConfiguration,
    ScanResult,
    Vulnerability,
)
from yardstick.label import find_labels_for_match, label_entry_matches_image_lineage
from yardstick.tool import get_tool


# AgainstLabels compares a scan result against a set of labels
@dataclass_json
@dataclass
class AgainstLabels:
    # basic identifying information
    config: ScanConfiguration = field(init=False)

    label_entries: list[LabelEntry]

    ###################
    # Relative to the matches from the result, which may not be complete

    # { label : [matches with the label] }
    # why not a set? we want to preserve the counts
    matches_by_label: dict[Label, list[Match]] = field(init=False)

    # { match_id : list(labels found for match) }
    # why not a set? we want to preserve the counts
    labels_by_match: dict[str, list[Label]] = field(init=False)
    label_entries_by_match: dict[str, list[LabelEntry]] = field(init=False)

    # matches that:
    # - do not have exactly one label
    # - have an Unclear label
    matches_with_indeterminate_labels: list[Match] = field(init=False)

    ##################
    # Relative to the aggregate of matches from results and MatchLabels

    false_positive_matches: list[Match] = field(
        init=False,
    )  # discovered matches in the result where there is a label that is NOT TP
    true_positive_matches: list[Match] = field(
        init=False,
    )  # discovered matches in the result where there is a TP label found
    false_negative_label_entries: set[LabelEntry] = field(
        init=False,
    )  # labels found that indicate TP, but not found as a match in the result

    result: InitVar[ScanResult]
    lineage: InitVar[list[str]]

    fuzzy_package_match: bool = field(default=False)

    compare_configuration: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self, result: ScanResult, lineage: list[str]):
        self.config = result.config

        if not self.compare_configuration:
            self.compare_configuration = {}

        self.compare_configuration["fuzzy_package_match"] = self.fuzzy_package_match

        self.matches_by_label = {
            Label.TruePositive: [],
            Label.FalsePositive: [],
            Label.Unclear: [],
        }
        self.labels_by_match = {}
        self.label_entries_by_match = {}
        self.false_negative_label_entries = {
            label
            for label in self.label_entries
            if label_entry_matches_image_lineage(label, result.config.image, lineage) and label.label == Label.TruePositive
        }
        self.true_positive_matches = []
        self.false_positive_matches = []
        self.matches_with_indeterminate_labels = []

        if result.matches is None:
            raise ValueError("no matches found in result")

        matched_true_positive_label_entries = set()

        for match in result.matches:
            # capture matches_by_label & labels_by_match ...
            label_entries = find_labels_for_match(
                result.config.image,
                match,
                self.label_entries,
                lineage=lineage,
                fuzzy_package_match=self.fuzzy_package_match,
            )

            for label_entry in label_entries:
                if label_entry.label == Label.TruePositive:
                    matched_true_positive_label_entries.add(label_entry)

            # remove any labels that have been paried with a result match
            self.false_negative_label_entries -= set(label_entries)

            self.label_entries_by_match[match.ID] = label_entries
            self.labels_by_match[match.ID] = [label_entry.label for label_entry in label_entries]
            labels_for_match = self.labels_by_match[match.ID]
            label_set = set(labels_for_match)

            for label in label_set:
                self.matches_by_label[label].append(match)

            # capture any matches that are in an indeterminate state...
            # note: it is important NOT to use a set here, as multiple instances of the same label will throw
            # the F1 score (e.g. [TP, TP ,TP] != [TP], only the latter is acceptable).
            # note: we're not considering unclear label to count as an indeterminate match
            # we want a definite f1 score in the quality gate without being overwhelmed by "disputed" unclear label
            if len(label_set) != 1:
                self.matches_with_indeterminate_labels.append(match)
                # capture TP & FP only beyond this point...
                continue

            if Label.TruePositive in label_set:
                self.true_positive_matches.append(match)
            elif Label.FalsePositive in label_set:
                self.false_positive_matches.append(match)

        # let's do one more pass regarding FNs. Since this is calculated based on the label entries we have,
        # and there may be multiple ways to represent the same vuln (ELSA-* and CVE-*), we need to ensure that
        # we remove any FNs that are actually TPs, but represented differently. This involves some guess work.
        self.false_negative_label_entries = prune_represented_fns(self.false_negative_label_entries, matched_true_positive_label_entries)

        self.summary = LabelComparisonSummary(result=result, comparison=self)

    def __str__(self):
        return str(self.summary)


def prune_represented_fns(
    false_negative_label_entries: set[LabelEntry],
    matched_true_positive_label_entries: set[LabelEntry],
) -> set[LabelEntry]:
    # for every FN, if there is a TP that is a subset of the FN, remove the FN

    remove_fns = set()
    evidence = set()
    for tp in matched_true_positive_label_entries:
        for fn in false_negative_label_entries:
            if tp.package != fn.package:
                continue
            if has_overlapping_vulnerability_id(tp, fn):
                remove_fns.add(fn)
                evidence.add((tp, fn))

    return false_negative_label_entries - remove_fns


def has_overlapping_vulnerability_id(tp: LabelEntry, fn: LabelEntry) -> bool:
    left_ids = {tp.vulnerability_id, tp.effective_cve}
    right_ids = {fn.vulnerability_id, fn.effective_cve}

    if "" in left_ids:
        left_ids.remove("")

    if "" in right_ids:
        right_ids.remove("")

    result = bool(left_ids & right_ids)

    return result


def _f1_score(tp, fp, fn):
    if tp + fp + fn == 0:
        return 0
    return tp / (tp + (0.5 * (fp + fn)))


@dataclass_json
@dataclass
class LabelComparisonSummary:
    # basic identifying information
    config: ScanConfiguration = field(init=False)

    total: int = field(init=False)
    indeterminate: int = field(init=False)
    indeterminate_percent: float = field(init=False)

    false_positives: int = field(init=False)
    true_positives: int = field(init=False)
    false_negatives: int = field(init=False)

    f1_score: float = field(init=False)
    f1_score_lower_confidence: float = field(init=False)
    f1_score_upper_confidence: float = field(init=False)
    f1_score_is_practicable: bool = field(init=False)

    # note: does NOT support ByMatchComponent (yet?)
    comparison: InitVar[AgainstLabels]
    result: InitVar[ScanResult] = None

    def __post_init__(self, comparison, result):
        self.config = comparison.config  # basic identifying information

        self.total = len(result.matches)
        self.indeterminate = len(comparison.matches_with_indeterminate_labels)
        self.indeterminate_percent = utils.safe_div(self.indeterminate, self.total) * 100
        self.true_positives = len(comparison.true_positive_matches)
        self.false_positives = len(comparison.false_positive_matches)
        self.false_negatives = len(comparison.false_negative_label_entries)

        self.f1_score = _f1_score(
            self.true_positives,
            self.false_positives,
            self.false_negatives,
        )
        self.f1_score_upper_confidence = _f1_score(
            self.true_positives + self.indeterminate,
            self.false_positives,
            self.false_negatives,
        )
        self.f1_score_lower_confidence = _f1_score(
            self.true_positives,
            self.false_positives + self.indeterminate,
            self.false_negatives,
        )

        self.f1_score_is_practicable = (self.f1_score_upper_confidence - self.f1_score_lower_confidence) <= 0.1  # is less than 10% variance

    def __str__(self):
        lines = []

        lines.append(f"Image: {self.config.image}")
        lines.append(f"Tool: {self.config.tool_name} @ {self.config.tool_version}")

        indent = "    "

        usable = ""
        if not self.f1_score_is_practicable:
            usable = "[IMPRACTICAL]"

        lines.append("Matches:")
        results_table = [
            ["True positive", self.true_positives],
            ["False positive", self.false_positives],
            ["False negative", self.false_negatives],
            [
                "Indeterminate",
                f"{self.indeterminate} ({self.indeterminate_percent:0.2f}%)",
            ],
            ["Total", self.total],
        ]

        lines.append(
            indent + tabulate(results_table, tablefmt="plain").replace("\n", "\n" + indent),
        )

        lines.append(f"F1 score        : {self.f1_score:0.2f}       {usable}")
        lines.append(
            f"(possible range): {self.f1_score_lower_confidence:0.2f}-{self.f1_score_upper_confidence:0.2f}",
        )

        return "\n".join(lines)


@dataclass(frozen=True)
class EquivalentMatch:
    vulnerability: Vulnerability
    package: Package
    matches: dict[str, list[Match]]


# ByPreservedMatch achieves the same effect as ByMatch, where set logic considers the relationships of
# packages and vulnerabilities, however, the difference is that all equivalent matches from each respective
# result source is preserved. This is useful in cases where you want to get the common and unique matches
# across multiple result sources, but are also interested into digging into specific elements in each
# record that are unrelated to data used in the set logic (e.g. you want to see the package IDs for each
# match for each record that is equivalent).
@dataclass
class ByPreservedMatch:
    results: Sequence[ScanResult]

    configs: dict[str, ScanConfiguration] = field(init=False)
    match_set: dict[str, set[Match]] = field(init=False)
    common: list[EquivalentMatch] = field(init=False)
    unique: dict[str, set[Match]] = field(init=False)

    summary: RelativeComparisonSummary = field(init=False)

    def __post_init__(self):
        self.configs = {result.ID: result.config for result in self.results}

        # create a set of matches discovered from each result
        self.match_set = {}
        for result in self.results:
            self.match_set[result.ID] = set(result.matches)

        # what set of matches are unique to each result?
        self.unique = {}
        for result in self.results:
            other_sets = [matches for res_id, matches in self.match_set.items() if res_id != result.ID]
            self.unique[result.ID] = self.match_set[result.ID].difference(*other_sets)

        # what set of matches were discovered across all results? Keep track of each result that matches
        # with any other result --they should be grouped together into a EquivalentMatch
        hashes: dict[int, list[tuple[str, Match]]] = collections.defaultdict(list)
        for result_id, matches in self.match_set.items():
            for match in matches:
                hashes[hash(match)].append((result_id, match))

        self.common = []
        for _, matches in hashes.items():
            # only capture matches as "common" if all of the results have a representative
            if len({result_id for result_id, _ in matches}) == len(self.results):
                match_group: dict[str, list[Match]] = collections.defaultdict(list)
                for result_id, match in matches:
                    match_group[result_id].append(match)

                ex = matches[0][1]
                vuln = Vulnerability(id=ex.vulnerability.id)
                pkg = Package(name=ex.package.name, version=ex.package.version)

                self.common.append(
                    EquivalentMatch(
                        vulnerability=vuln,
                        package=pkg,
                        matches=dict(match_group),
                    ),
                )

        self.summary = RelativeComparisonSummary(comparison=self, results=self.results)

    def __str__(self):
        return str(self.summary)


# ByMatch takes the set() approach, coupling the hash()/eq() logic for Match objects and all children
# and evaluating the results based on python set operations. The upside to this comparator is the ease
# of implementation and groking (common = A intersect B , unique[id] = all other packages not in common),
# the downside is that the original objects from each result group are not captured, only a SINGLE result
# is captured, and which element is selected is arbitrary. This is useful if you are trying to collect
# stats about these groups, and common information (such as package name, version, and vulnerability ID).
# This is not useful if you need the original records from all result sets which are equivalent, in that
# case use ByPreservedMatch.
@dataclass
class ByMatch:
    results: Sequence[ScanResult]

    configs: dict[str, ScanConfiguration] = field(init=False)
    match_set: dict[str, set[Match]] = field(init=False)
    common: set[Match] = field(init=False)
    unique: dict[str, set[Match]] = field(init=False)

    summary: RelativeComparisonSummary = field(init=False)

    def __post_init__(self):
        self.configs = {result.ID: result.config for result in self.results}

        # create a set of matches discovered from each result
        self.match_set = {}
        for result in self.results:
            self.match_set[result.ID] = set(result.matches)

        # what set of matches were discovered across all results?
        self.common = set.intersection(*self.match_set.values())

        # what set of matches are unique to each result?
        self.unique = {}
        for result in self.results:
            other_sets = [matches for res_id, matches in self.match_set.items() if res_id != result.ID]
            self.unique[result.ID] = self.match_set[result.ID].difference(*other_sets)

        self.summary = RelativeComparisonSummary(comparison=self, results=self.results)

    def __str__(self):
        return str(self.summary)


# ByVulnerability looks solely at at matches vulnerability for comparison. This alone isn't an accurate comparison,
# however, if you have two tools that have different package name/version extraction algorithms that result in
# slightly different values, this comparison is a fair proxy for answering the question "did both tools find the
# same CVEs or different CVEs for the same image".
@dataclass
class ByVulnerability:
    results: Sequence[ScanResult]

    configs: dict[str, ScanConfiguration] = field(init=False)
    vulnerabilities: dict[str, list[Vulnerability]] = field(init=False)
    vulnerability_set_by_result_id: dict[str, set[Vulnerability]] = field(init=False)
    common: set[Vulnerability] = field(init=False)
    unique: dict[str, set[Vulnerability]] = field(init=False)

    def __post_init__(self):
        self.configs = {result.ID: result.config for result in self.results}

        # create a set of vulnerabilities discovered from each result
        self.vulnerability_set_by_result_id = {}
        self.vulnerabilities = {}
        for result in self.results:
            self.vulnerabilities[result.ID] = [m.vulnerability.id for m in result.matches]
            self.vulnerability_set_by_result_id[result.ID] = set(
                self.vulnerabilities[result.ID],
            )

        # what set of vulnerabilities were discovered across all results?
        self.common = set.intersection(*self.vulnerability_set_by_result_id.values())

        # what set of vulnerabilities are unique to each result?
        self.unique = {}
        for result in self.results:
            other_sets = [vulnerabilities for res_id, vulnerabilities in self.vulnerability_set_by_result_id.items() if res_id != result.ID]
            self.unique[result.ID] = self.vulnerability_set_by_result_id[result.ID].difference(*other_sets)


# ByPackage looks solely at at matches package info for comparison. This alone isn't an accurate comparison,
# however, if you have two tools that have different package name/version extraction algorithms that result in
# slightly different values, this comparison is a fair proxy for answering the question "did both tools find the
# same packages or different packages for the same image (with respect to match findings)".
@dataclass
class ByPackage:
    results: Sequence[ScanResult]

    configs: dict[str, ScanConfiguration] = field(init=False)
    packages: dict[str, list[Package]] = field(init=False)
    package_set_by_result_id: dict[str, set[Package]] = field(init=False)
    common: set[Package] = field(init=False)
    unique: dict[str, set[Package]] = field(init=False)

    def __post_init__(self):
        self.configs = {result.ID: result.config for result in self.results}

        # create a set of packages discovered from each result
        self.package_set_by_result_id = {}
        self.packages = {}
        for result in self.results:
            self.packages[result.ID] = [m.package for m in result.matches]
            self.package_set_by_result_id[result.ID] = set(self.packages[result.ID])

        # what set of packages were discovered across all results?
        self.common = set.intersection(*self.package_set_by_result_id.values())

        # what set of packages are unique to each result?
        self.unique = {}
        for result in self.results:
            other_sets = [packages for res_id, packages in self.package_set_by_result_id.items() if res_id != result.ID]
            self.unique[result.ID] = self.package_set_by_result_id[result.ID].difference(*other_sets)


@dataclass_json
@dataclass
class ComponentSummary:
    # count of matches found in all result sets
    deduplicated_common_count: int = field(init=False)

    # count of all matches for a result set
    total_count: dict[str, int] = field(init=False)

    # count of all matches that are unique within each result set
    deduplicated_total_count: dict[str, int] = field(init=False)

    # matches found only in each result set
    deduplicated_unique_count: dict[str, int] = field(init=False)

    # this is the same as the F1 score, but an F1 score is typically relative to a ground truth, where as dice
    # similarity does not imply this... it is a measure of the similarity of two samples without an assumption that one
    # of the samples is a ground truth (a small but important semantic difference).
    dice_similarity_coefficient: float = field(init=False)

    # elements needed to compute the summary
    common: InitVar[Sequence[Any]]
    originals: InitVar[dict[str, Sequence[Any]]]
    uniques: InitVar[dict[str, Sequence[Any]]]

    def __post_init__(
        self,
        common: Sequence[Any],
        originals: dict[str, Sequence[Any]],
        uniques: dict[str, Sequence[Any]],
    ):
        if originals.keys() != uniques.keys():
            raise RuntimeError(
                f"mismatched component elements: {originals.keys()} {uniques.keys()}",
            )
        self.deduplicated_common_count = len(common)
        self.total_count = {}
        self.deduplicated_total_count = {}
        self.deduplicated_unique_count = {}
        for parent_id, items in originals.items():
            self.total_count[parent_id] = len(items)
            self.deduplicated_total_count[parent_id] = len(set(items))
            self.deduplicated_unique_count[parent_id] = len(set(uniques[parent_id]))

        # | A intersect B | / | A union B | = | A intersect B | / ((| A | + | B |) - | A intersect B |)
        # note: this is done on the count of SETs of matches... we are ignore duplicates here
        # why? because the common count is a set, so we cannot use non-set counts for the two data groups.
        jaccard_similarity = utils.safe_div(
            self.deduplicated_common_count,
            sum(self.deduplicated_total_count.values()) - self.deduplicated_common_count,
        )
        # now we can easily compute the dice similarity...
        self.dice_similarity_coefficient = (2 * jaccard_similarity) / (jaccard_similarity + 1)


@dataclass_json
@dataclass
class RelativeComparisonSummary:
    # basic identifying information
    configs: dict[str, ScanConfiguration] = field(init=False)

    match_component: ComponentSummary = field(init=False)
    vulnerability_component: ComponentSummary = field(init=False)
    package_component: ComponentSummary = field(init=False)

    # note: does NOT support ByMatchComponent (yet?)
    comparison: InitVar[ByPreservedMatch | ByMatch]
    results: InitVar[Sequence[ScanResult]] = None

    def __post_init__(self, comparison, results):
        self.configs = comparison.configs  # basic identifying information
        self.match_component = ComponentSummary(
            common=comparison.common,
            # note: this is the original LIST, not de-duplicated set
            originals={result.ID: result.matches for result in results},
            uniques=comparison.unique,
        )

        # once you see the dice similarity, you may want to see other kinds of comparisons to determine why the similarity
        # is what it is. This is where a comparison by sub-components could help.
        comparison_by_vulnerabilities = ByVulnerability(results)
        self.vulnerability_component = ComponentSummary(
            common=comparison_by_vulnerabilities.common,
            # note: this is the original LIST, not de-duplicated set
            originals=comparison_by_vulnerabilities.vulnerabilities,
            uniques=comparison_by_vulnerabilities.unique,
        )

        comparison_by_packages = ByPackage(results)
        self.package_component = ComponentSummary(
            common=comparison_by_packages.common,
            # note: this is the original LIST, not de-duplicated set
            originals=comparison_by_packages.packages,
            uniques=comparison_by_packages.unique,
        )

    def __str__(self):
        lines = []

        if len({config.image for _, config in self.configs.items()}) == 1:
            lines.append(f"Image: {next(iter(self.configs.values())).image}")

        indent = "    "

        results_table: list[list[Any]] = [
            [
                "TOOL",
                "MATCHES",
                "",
                "PACKAGES",
                "VULNERABILITES",
            ],
        ]
        for result_id, config in self.configs.items():
            results_table.append(
                [
                    config.tool,
                    self.match_component.total_count[result_id],
                    f"({self.match_component.deduplicated_total_count[result_id]} unique)",
                    self.package_component.deduplicated_total_count[result_id],
                    self.vulnerability_component.deduplicated_total_count[result_id],
                ],
            )

        lines.append(
            indent + tabulate(results_table, tablefmt="plain").replace("\n", "\n" + indent) + "\n",
        )

        lines.append("Comparison Results (deduplicated): ")

        table: list[list[Any]] = [
            [
                "TOOL",
                "MATCHES",
                "PACKAGES",
                "VULNERABILITES",
            ],
        ]
        table.append(
            [
                "(common to all)",
                f"{self.match_component.deduplicated_common_count}",
                f"{self.package_component.deduplicated_common_count}",
                f"{self.vulnerability_component.deduplicated_common_count}",
            ],
        )

        for result_id, config in self.configs.items():
            table.append(
                [
                    f"unique to {config.tool}",
                    self.match_component.deduplicated_unique_count[result_id],
                    self.package_component.deduplicated_unique_count[result_id],
                    self.vulnerability_component.deduplicated_unique_count[result_id],
                ],
            )

        lines.append(
            indent + tabulate(table, tablefmt="plain").replace("\n", "\n" + indent) + "\n",
        )

        lines.append(
            f"Package similarity       : {self.package_component.dice_similarity_coefficient:0.2f}",
        )
        lines.append(
            f"Vulnerability similarity : {self.vulnerability_component.dice_similarity_coefficient:0.2f}",
        )
        lines.append(
            f"Match similarity         : {self.match_component.dice_similarity_coefficient:0.2f}",
        )
        return "\n".join(lines)


@dataclass_json
@dataclass(frozen=True)
class ImageToolLabelStats:
    configs: list[ScanConfiguration]
    compare_configs: list[dict[str, str]] = field(default_factory=list)
    indeterminate: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    indeterminate_percent: dict[str, dict[str, float]] = field(
        default_factory=lambda: collections.defaultdict(dict),
    )
    true_positives: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    false_positives: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    false_negatives: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    f1_scores: dict[str, dict[str, float]] = field(
        default_factory=lambda: collections.defaultdict(dict),
    )
    f1_score_ranges: dict[str, dict[str, tuple[float, float]]] = field(
        default_factory=lambda: collections.defaultdict(dict),
    )

    @property
    def images(self):
        return sorted(self.f1_scores.keys())

    @property
    def tools(self):
        tools = set()
        for image in self.f1_scores:
            tools.update(self.f1_scores[image].keys())
        return sorted(tools)

    @staticmethod
    def new(comparisons: list[AgainstLabels]) -> ImageToolLabelStats:
        configs = [comp.config for comp in comparisons]

        compare_configs = [comp.compare_configuration for comp in comparisons]

        indeterminate: dict[str, dict[str, int]] = collections.defaultdict(
            lambda: collections.defaultdict(int),
        )
        indeterminate_percent: dict[str, dict[str, float]] = collections.defaultdict(
            dict,
        )
        true_positives: dict[str, dict[str, int]] = collections.defaultdict(
            lambda: collections.defaultdict(int),
        )
        false_positives: dict[str, dict[str, int]] = collections.defaultdict(
            lambda: collections.defaultdict(int),
        )
        false_negatives: dict[str, dict[str, int]] = collections.defaultdict(
            lambda: collections.defaultdict(int),
        )
        f1_scores: dict[str, dict[str, float]] = collections.defaultdict(dict)
        f1_score_ranges: dict[
            str,
            dict[str, tuple[float, float]],
        ] = collections.defaultdict(dict)

        for comp in comparisons:
            image = comp.config.image
            tool = comp.config.tool

            if comp.summary.f1_score_is_practicable:
                f1_scores[image][tool] = comp.summary.f1_score
            else:
                f1_scores[image][tool] = -1

            if comp.summary.f1_score_lower_confidence != 0 and comp.summary.f1_score_upper_confidence != 0:
                f1_score_ranges[image][tool] = (
                    comp.summary.f1_score_lower_confidence,
                    comp.summary.f1_score_upper_confidence,
                )

            false_negatives[image][tool] = comp.summary.false_negatives
            true_positives[image][tool] = comp.summary.true_positives
            false_positives[image][tool] = comp.summary.false_positives
            indeterminate[image][tool] = comp.summary.indeterminate
            indeterminate_percent[image][tool] = utils.safe_div(comp.summary.indeterminate, comp.summary.total) * 100.0

        return ImageToolLabelStats(
            configs=configs,
            compare_configs=compare_configs,
            indeterminate=indeterminate,
            indeterminate_percent=indeterminate_percent,
            true_positives=true_positives,
            false_positives=false_positives,
            false_negatives=false_negatives,
            f1_scores=f1_scores,
            f1_score_ranges=f1_score_ranges,
        )


def of_results_against_label(
    *results: ScanResult,
    label_entries,
    fuzzy_package_match: bool = False,
    compare_configuration: dict[str, Any] | None = None,
) -> tuple[dict[str, AgainstLabels], ImageToolLabelStats]:
    logging.debug("starting comparison against labels")

    comparisons_by_result_id = {}

    if not compare_configuration:
        compare_configuration = {}

    comparisons = []
    for result in results:
        logging.debug(
            f"comparing labels for image={result.config.image} tool={result.config.tool}",
        )

        lineage = store.image_lineage.get(image=result.config.image)
        comp = AgainstLabels(
            result=result,
            label_entries=label_entries,
            lineage=lineage,
            fuzzy_package_match=fuzzy_package_match,
            compare_configuration=compare_configuration,
        )
        comparisons_by_result_id[result.ID] = comp
        comparisons.append(comp)

    stats_by_image_tool_pair = ImageToolLabelStats.new(comparisons)

    return comparisons_by_result_id, stats_by_image_tool_pair


@dataclass
class ToolLabelStatsByEcosystem:
    tps_by_tool_by_ecosystem: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    fps_by_tool_by_ecosystem: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    precision_by_tool_by_ecosystem: dict[str, dict[str, int]] = field(
        default_factory=lambda: collections.defaultdict(
            lambda: collections.defaultdict(int),
        ),
    )
    tools: list[str] = field(default_factory=list)
    ecosystems: list[str] = field(default_factory=list)

    @staticmethod
    def new(  # noqa: C901
        comparisons_by_result_id_by_image: dict[str, list[dict[str, AgainstLabels]]],
    ) -> ToolLabelStatsByEcosystem:
        stats = ToolLabelStatsByEcosystem()

        def normalize_ecosystem(e: str) -> str:
            e = e.lower().replace("-", "_")
            if e in ("java_archive", "jenkins_plugin"):
                return "java"
            return e

        def ecosystem(match: Match) -> str:
            if not match.config:
                return "unknown"
            t = get_tool(match.config.tool_name)
            if not t:
                return "unknown"
            # why ignore the type? combining staticmethod and abstract method leads to false positive
            package_type = t.parse_package_type(match.fullentry)  # type: ignore[union-attr, arg-type]
            return normalize_ecosystem(package_type)

        tools = set()
        ecosystems = set()
        for _, comparisons in comparisons_by_result_id_by_image.items():
            for comparisons_by_result_id in comparisons:
                for _, comp in comparisons_by_result_id.items():
                    tool = comp.config.tool
                    tools.add(tool)

                    for tp in comp.true_positive_matches:
                        e = ecosystem(tp)
                        ecosystems.add(e)
                        stats.tps_by_tool_by_ecosystem[tool][e] += 1

                    for fp in comp.false_positive_matches:
                        e = ecosystem(fp)
                        ecosystems.add(e)
                        stats.fps_by_tool_by_ecosystem[tool][e] += 1

        stats.tools = sorted(tools)
        stats.ecosystems = sorted(ecosystems)

        # calculate precision
        for tool in tools:
            for e in ecosystems:
                stats.precision_by_tool_by_ecosystem[tool][e] = utils.safe_div(
                    stats.tps_by_tool_by_ecosystem[tool][e],
                    stats.tps_by_tool_by_ecosystem[tool][e] + stats.fps_by_tool_by_ecosystem[tool][e],
                )

        return stats


def of_results_against_label_by_ecosystem(
    results_by_image: dict[str, list[ScanResult]],
    label_entries,
    fuzzy_package_match: bool = False,
) -> ToolLabelStatsByEcosystem:
    comparisons_by_result_id_by_image = collections.defaultdict(list)
    for image, results in results_by_image.items():
        comparisons_by_result_id, _ = of_results_against_label(
            *results,
            label_entries=label_entries,
            fuzzy_package_match=fuzzy_package_match,
        )
        comparisons_by_result_id_by_image[image].append(comparisons_by_result_id)

    return ToolLabelStatsByEcosystem.new(comparisons_by_result_id_by_image)



import logging
import os
import sqlite3
import subprocess
import sys
import threading
from contextlib import closing
from typing import Optional


class GrypeDBManager:
    enabled: bool
    message: str
    db_location: Optional[str]
    connections: dict[int, sqlite3.Connection]

    def __init__(self, db_location: Optional[str] = None):
        self.enabled = False
        self.message = ""
        self.db_location = db_location
        self.connections = {}

        if self.db_location:
            try:
                self.connect()
            except:  # noqa: E722
                self.db_location = None
                logging.error(
                    f"unable to open grype DB at {self.db_location}. Falling back to system grype DB.",
                )

        if not self.db_location:
            self.set_db_to_system_grype_db()

        if self.db_location:
            self.enabled = True

    def close(self):
        for conn in self.connections.values():
            conn.close()
        self.connections = {}

    def set_db_to_system_grype_db(self):
        try:
            logging.debug("using system grype DB...")
            out = subprocess.check_output(
                ["grype", "db", "status"],
            ).decode(
                sys.stdout.encoding,
            )
            for line in out.split("\n"):
                if line.startswith("Path:"):
                    self.db_location = line.removeprefix("Path:").strip().removesuffix("vulnerability.db")
        except Exception as e:
            self.message = str(e)
            logging.error("unable to open grype DB %s", e)

    def connect(self):
        # sqlite3 is not thread safe, so we need to create a connection per thread
        tid = threading.get_ident()
        if tid in self.connections:
            return self.connections[tid]

        conn = sqlite3.connect(os.path.join(self.db_location, "vulnerability.db"))
        self.connections[tid] = conn
        return conn

    def get_upstream_vulnerability(self, vuln_id: str) -> Optional[str]:
        with closing(self.connect().cursor()) as cur:
            cur.execute(
                "SELECT alias FROM vulnerability_aliases where name == ? ORDER BY alias ASC LIMIT 1;",
                (vuln_id,),
            )
            vulnerability_info = cur.fetchall()

        for info in vulnerability_info:
            if info and len(info) > 0 and info[0]:
                return info[0]

        return None

    def get_vuln_description(self, vuln_id: str, provider: Optional[str] = None) -> str:
        with closing(self.connect().cursor()) as cur:
            if provider:
                cur.execute(
                    """
                    SELECT
                        json_extract(b.value, '$.description') description
                    FROM
                        vulnerability_handles vh
                        INNER JOIN blobs b
                            ON b.id = vh.blob_id
                    WHERE
                        vh.name == ?
                        AND vh.provider_id == ?
                    ;
                    """,
                    (vuln_id, provider),
                )
            else:
                cur.execute(
                    """
                    SELECT
                        json_extract(b.value, '$.description') description
                    FROM
                        vulnerability_handles vh
                        INNER JOIN blobs b
                            ON b.id = vh.blob_id
                    WHERE
                        vh.name == ?
                    ;
                    """,
                    (vuln_id,),
                )
            results = cur.fetchall()

        for result in results:
            if result and len(result) > 0 and result[0]:
                return result[0]
        return ""

    # get vulnerability description of a vulnerability and all its related vulnerabilities
    def get_all_vulnerability_descriptions(self, vuln_id: str) -> str:
        if not self.db_location:
            return ""

        upstream = self.get_upstream_vulnerability(vuln_id)

        message = ""

        if upstream and upstream != vuln_id:
            provider = None

            if vuln_id.lower().startswith("cve-"):
                provider = "nvd"
            elif vuln_id.lower().startswith("ghsa-"):
                provider = "github"

            vuln_desc = self.get_vuln_description(upstream, provider=provider)
            if vuln_desc:
                message += f"Upstream Vulnerability: {upstream}\n{vuln_desc}\n\n"

        vuln_desc = self.get_vuln_description(vuln_id)
        if vuln_desc:
            message += f"Vulnerability: {vuln_id}\n{vuln_desc}\n"

        return message


_instance = None
_raise_on_failure = False


def raise_on_failure(value: bool):
    global _raise_on_failure  # noqa: PLW0603
    _raise_on_failure = value


def use(location: str):
    global _instance  # noqa: PLW0603

    if _instance:
        _instance.close()

    _instance = GrypeDBManager(location)


def normalize_to_cve(vuln_id: str):
    global _instance  # noqa: PLW0603
    if vuln_id.lower().startswith("cve-"):
        return vuln_id

    try:
        if not _instance:
            _instance = GrypeDBManager()

        upstream = _instance.get_upstream_vulnerability(vuln_id)
    except:  # noqa: E722
        if _raise_on_failure:
            raise
        return vuln_id

    if upstream and upstream.lower().startswith("cve-"):
        return upstream

    # unable to normalize, return the original
    return vuln_id



import json
from typing import Callable, Dict, Optional

import pygments
from prompt_toolkit import PromptSession, print_formatted_text
from prompt_toolkit.completion import (
    CompleteEvent,
    Completer,
    Completion,
    merge_completers,
)
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import HTML, PygmentsTokens
from prompt_toolkit.styles import style_from_pygments_cls
from prompt_toolkit.validation import ValidationError, Validator
from pygments.lexers.data import JsonLexer
from pygments.styles.monokai import MonokaiStyle

from yardstick import artifact
from yardstick.tool import get_tool


def display_match_table_row(match: artifact.Match) -> str:
    if not match.config:
        return ""
    t = get_tool(match.config.tool_name)
    package_type = t.parse_package_type(match.fullentry)  # type: ignore[union-attr]
    return f"{match.vulnerability.id:20} {match.package.name}@{match.package.version} {package_type}"


def display_match(match: artifact.Match) -> str:
    if not match.config:
        return ""
    t = get_tool(match.config.tool_name)
    package_type = t.parse_package_type(match.fullentry)  # type: ignore[union-attr]
    pkg = f"{match.package.name}@{match.package.version}"
    return f"match vuln='{match.vulnerability.id}', cve='{match.vulnerability.cve_id}', package='{pkg}', type='{package_type}', id='{match.ID}'"


class MatchCollection:
    def __init__(self, result: artifact.ScanResult):
        self.result = result
        if not self.result.matches:
            raise ValueError("no matches provided")

        self.match_display_text_by_id = {m.ID: display_match(m) for m in self.result.matches}
        self.match_by_id = {m.ID: m for m in self.result.matches}
        self.match_id_by_display_text = {v: k for k, v in self.match_display_text_by_id.items()}

    def has_display_text(self, text):
        return text in self.match_display_text_by_id.values()

    def get_match_display_text(self, match):
        return self.match_display_text_by_id[match.ID]

    def get_match(self, text) -> Optional[artifact.Match]:
        match_id = self.match_id_by_display_text.get(text, None)
        if match_id:
            return self.match_by_id.get(match_id, None)
        return None

    def get_matches(self, filter_text=None):
        if filter_text:
            filter_text = filter_text.lower()

            def condition(match: artifact.Match) -> bool:
                if not match.config:
                    return False
                t = get_tool(match.config.tool_name)
                package_type = t.parse_package_type(match.fullentry)  # type: ignore[union-attr]

                return (
                    filter_text in match.vulnerability.id.lower()
                    or filter_text in match.package.name.lower()
                    or filter_text in match.package.version.lower()
                    or filter_text in package_type
                )

        else:

            def condition(
                match: artifact.Match,
            ) -> bool:
                return True

        return [match for match in sorted(self.result.matches) if condition(match)]


class ResultCompleter(Completer):
    def __init__(self, matches: MatchCollection):
        self.matches = matches

    def get_completions(self, document: Document, complete_event: CompleteEvent):
        matches = self.matches.get_matches(filter_text=document.text.strip())
        if document.text.lower().startswith("match"):
            matches += self.matches.get_matches(
                filter_text=document.text.lstrip("match").strip(),
            )

        for match in matches:
            yield Completion(
                self.matches.get_match_display_text(match),
                start_position=-len(document.text_before_cursor),
                display=display_match_table_row(match),
                display_meta="match",
            )


class ExploreValidator(Validator):
    def __init__(self, completers, matches):
        self.completers = completers
        self.matches = matches

    def validate(self, document: Document):
        # did we find a match or command? then error out
        if not self.completers.get_completions(
            document,
            None,
        ) and not self.matches.has_display_text(document.text):
            raise ValidationError(message="Not matches found")


class Executor(Completer):
    def __init__(self, matches: MatchCollection):
        self.matches = matches
        self.commands: Dict[str, Callable] = {
            "list": self.list,
            "help": self.help,
            "match": self.match,
        }
        self.display: Dict[str, HTML] = {
            "list": HTML("<b>list</b> [optional filter text]"),
            "help": HTML("<b>help</b>"),
            "match": HTML(
                "<b>match</b> <i>vuln</i>=str <i>package</i>=str <i>id</i>=str",
            ),
        }
        self.command_descriptions: Dict[str, Optional[str]] = {cmd: fn.__doc__ for cmd, fn in self.commands.items()}

    def get_completions(self, document: Document, complete_event: CompleteEvent):
        text = document.text.lower()
        for name in self.commands:
            if text.startswith(name) or name.startswith(text):
                yield Completion(
                    name,
                    start_position=-len(document.text_before_cursor),
                    display=self.display[name],
                    display_meta="command",
                )

    def execute(self, text):
        cmd = self.commands.get(text.split(" ").pop(0), None)
        if cmd:
            cmd(text)
        else:
            print(f"could not parse input '{text}'")

    def match(self, text: str):
        """
        show original json entry for a given match (prompt matches partial CVE, package name, and package version))
        """
        match = self.matches.get_match(text)
        if match:
            json_str = json.dumps(match.fullentry, indent=2)
            tokens = list(pygments.lex(json_str, lexer=JsonLexer()))
            style = style_from_pygments_cls(pygments_style_cls=MonokaiStyle)
            print_formatted_text(PygmentsTokens(tokens), style=style)

    def list(self, text: Optional[str] = None):  # noqa: A003
        """
        list all vulnerability matches (accepts optional filter argument)
        """
        if text:
            text = text.lstrip("list").strip()

        matches = [f"{num + 1!s:3} | " + display_match(match) for num, match in enumerate(sorted(self.matches.get_matches(text)))]
        print_formatted_text("\n".join(matches))

    def help(self, _: Optional[str] = None):  # noqa: A003
        """
        show available commands
        """
        messages = []
        for k in sorted(self.commands):
            description = ""
            command_description = self.command_descriptions.get(k, None)
            if command_description:
                description = ": " + command_description
            messages.append("   <b>" + k + "</b>" + description)
        print_formatted_text(HTML("\n".join(["Commands:", *messages])))


def bottom_toolbar(result: artifact.ScanResult):
    def _render():
        info = f"<b>{result.config.image} {result.config.tool_name}@{result.config.tool_version}</b>"
        stats = [f"[matches <style bg='ansired'>{len(result.matches)}</style>]"]

        unique = len(set(result.matches))
        if unique != len(result.matches):
            stats += [f"[unique matches <style bg='ansired'>{len(set())}</style>]"]

        return HTML(f"{info} {' '.join(stats)}")

    return _render


def run(result: artifact.ScanResult):
    matches = MatchCollection(result)
    executor = Executor(matches)
    completer = ResultCompleter(matches)
    all_completers = merge_completers((completer, executor))
    vaidator = ExploreValidator(all_completers, matches)
    config = {
        "bottom_toolbar": bottom_toolbar(result),
        # mouse_support=True, # with mouse support, scrolling is disabled
        "completer": all_completers,
        "validator": vaidator,
    }
    session: PromptSession = PromptSession()

    executor.help()
    while True:
        text = session.prompt("> ", **config)
        executor.execute(text)



import atexit
import functools
import json
import logging
import io
import os
import re
import shlex
import shutil
import subprocess
import sys
import tarfile
import tempfile
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import git
import xxhash
import zstandard as zstd

from yardstick import artifact, utils
from yardstick.tool.vulnerability_scanner import VulnerabilityScanner
from yardstick.utils import github


@dataclass(frozen=False)
class GrypeProfile:
    name: Optional[str] = None
    config_path: Optional[str] = None


class Grype(VulnerabilityScanner):
    def __init__(  # noqa: PLR0913
        self,
        path: str,
        db_identity: str,
        version_detail: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        profile: Optional[GrypeProfile] = None,
        **kwargs,
    ):
        if not profile:
            profile = GrypeProfile()
        self.profile = profile
        self.db_identity = db_identity
        self.path = path
        self._env = env
        if version_detail:
            self.version_detail = version_detail
            if self.profile and self.profile.name:
                self.version_detail += f"+profile={self.profile.name}"

    @staticmethod
    @functools.cache
    def latest_version_from_github():
        return github.get_latest_release_version(project="grype")

    @staticmethod
    def _install_from_installer(
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ) -> "Grype":
        logging.debug(f"installing grype version={version!r} from installer")
        tool_exists = False

        if not use_cache and path:
            shutil.rmtree(path)

        if not path:
            path = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, path)

        if os.path.exists(os.path.join(path, "grype")):
            tool_exists = True

        if not tool_exists:
            subprocess.check_call(
                [
                    "sh",
                    "-c",
                    f"curl -sSfL https://raw.githubusercontent.com/anchore/grype/main/install.sh | sh -s -- -b {path} {version}",
                ],
            )

            # note on S103 exception: grype must be executable and it is not valid to assume that
            # a single user or group should be restricted to using it. 0755 is the default
            # permission used for system bin/* contents.
            os.chmod(f"{path}/grype", 0o755)  # noqa: S103
        else:
            logging.debug(f"using existing grype installation {path!r}")

        return Grype(path=path, version_detail=version, **kwargs)

    @classmethod
    def _install_from_git(  # noqa: PLR0912
        cls,
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ) -> "Grype":
        logging.debug(f"installing grype version={version!r} from git")
        tool_exists = False

        repo_url = "github.com/anchore/grype"
        if version.startswith("github.com"):
            repo_url = version
            if "@" in version:
                version = version.split("@")[1]
                repo_url = repo_url.split("@")[0]
            else:
                version = "main"

        if not use_cache and path:
            shutil.rmtree(path, ignore_errors=True)

        if not path:
            path = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, path)

        # grab the latest source code into a local directory
        repo_path = os.path.join(path, "source")

        if os.path.exists(repo_path):
            logging.debug(f"found existing grype repo at {repo_path!r}")
            # use existing source
            try:
                repo = git.Repo(repo_path)
            except:
                logging.error(f"failed to open existing grype repo at {repo_path!r}")
                raise
        else:
            logging.debug(f"cloning the grype git repo: {repo_url!r}")
            # clone the repo
            os.makedirs(repo_path)
            repo = git.Repo.clone_from(f"https://{repo_url}.git", repo_path)

        # checkout the ref in question
        repo.git.fetch("origin", version)
        repo.git.checkout(version)

        # get userful ref
        description = repo.git.describe("--tags", "--always", "--long")
        path = os.path.join(path, "git_install", description)

        logging.debug(f"found grype git description={description!r}")

        if os.path.exists(os.path.join(path, "grype")):
            tool_exists = True
        else:
            os.makedirs(path)

        abspath = os.path.abspath(path)
        if not tool_exists:
            logging.debug(f"installing grype to {abspath!r}")
            cls._run_go_build(
                abs_install_dir=abspath,
                repo_path=repo_path,
                description=description,
                binpath=path,
            )
        else:
            logging.debug(f"using existing grype installation {abspath!r}")

        return Grype(path=path, version_detail=description, **kwargs)

    @classmethod
    def _install_from_path(
        cls,
        path: Optional[str],
        src_path: str,
        use_cache: Optional[bool] = True,
        **kwargs,
    ) -> "Grype":
        if not use_cache and path:
            shutil.rmtree(path, ignore_errors=True)

        if not path:
            path = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, path)

        # get the description and head ref from the repo
        src_repo_path = os.path.abspath(os.path.expanduser(src_path))
        build_version = utils.local_build_version_suffix(src_repo_path)

        logging.debug(f"installing grype from path={src_repo_path!r} to path={path!r}")

        dest_path = os.path.join(
            path.replace("path:", ""),
            build_version,
            "local_install",
        )
        os.makedirs(dest_path, exist_ok=True)

        cls._run_go_build(
            abs_install_dir=os.path.abspath(dest_path),
            description=f"{path}:{build_version}",
            repo_path=src_repo_path,
            binpath=dest_path,
        )

        return Grype(path=dest_path, **kwargs)

    @staticmethod
    def _run_go_build(
        abs_install_dir: str,
        repo_path: str,
        description: str,
        binpath: str,
        version_ref: str = "github.com/anchore/grype/internal/version.version",
    ):
        logging.debug(f"installing grype via build to {abs_install_dir!r}")

        main_pkg_path = "./cmd/grype"
        if not os.path.exists(os.path.join(repo_path, "cmd", "grype", "main.go")):
            # support legacy installations, when the main.go was in the root of the repo
            main_pkg_path = "."

        c = f"go build -ldflags \"-w -s -extldflags '-static' -X {version_ref}={description}\" -o {abs_install_dir} {main_pkg_path}"
        logging.debug(f"running {c!r}")

        e = {"GOBIN": abs_install_dir, "CGO_ENABLED": "0"}
        e.update(os.environ)

        subprocess.check_call(
            shlex.split(c),
            stdout=sys.stdout,
            stderr=sys.stderr,
            cwd=repo_path,
            env=e,
        )

        # note on S103 exception: grype must be executable and it is not valid to assume that
        # a single user or group should be restricted to using it. 0755 is the default
        # permission used for system bin/* contents.
        os.chmod(f"{binpath}/grype", 0o755)  # noqa: S103

    @classmethod
    def install(  # noqa: C901, PLR0913
        cls,
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        update_db: bool = True,
        db_import_path=None,
        profile: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> "Grype":
        original_version = version
        specified_db = "+import-db=" in version
        db_identity = "oss"
        if specified_db:
            # note: doesn't allow for additional version modifiers
            fields = version.split("+import-db=")
            db_import_path = fields[1]
            version = fields[0]
            update_db = False
            db_identity = get_import_checksum(db_import_path)

        logging.debug(
            f"parsed import-db={db_import_path!r} from version={original_version!r} new version={version!r}",
        )
        grype_profile = GrypeProfile(**profile) if profile else GrypeProfile()

        if version == "latest":
            version = cls.latest_version_from_github()
            if path:
                path = os.path.join(os.path.dirname(path), version)
            logging.info(f"latest grype release found is {version}")

        # check if the version is a semver...
        if re.match(
            r"^v(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$",
            version,
        ):
            tool_obj = cls._install_from_installer(
                version=version,
                path=path,
                use_cache=use_cache,
                profile=grype_profile,
                db_identity=db_identity,
                **kwargs,
            )
        elif version.startswith("path:"):
            tool_obj = cls._install_from_path(
                path=path,
                src_path=version.removeprefix("path:"),
                version=version.removeprefix("path:"),
                use_cache=use_cache,
                db_identity=db_identity,
                profile=grype_profile,
                **kwargs,
            )
        else:
            tool_obj = cls._install_from_git(
                version=version,
                path=path,
                db_identity=db_identity,
                use_cache=use_cache,
                profile=grype_profile,
                **kwargs,
            )

        # always update the DB, raise exception on failure
        if db_import_path:
            if os.path.exists(tool_obj.db_root):
                logging.info(f"using existing (custom) db from {tool_obj.db_root!r}")
            else:
                logging.info(f"importing given (custom) db from {db_import_path!r}")
                tool_obj.run("db", "import", db_import_path)
        elif update_db:
            logging.debug("updating db from OSS")
            tool_obj.run("db", "update", "-vv")

        return tool_obj

    @property
    def db_root(self) -> str:
        return os.path.join(self.path, "db", self.db_identity)

    def env(self, override=None):
        env = os.environ.copy()
        env["GRYPE_CHECK_FOR_APP_UPDATE"] = "false"
        env["GRYPE_DB_VALIDATE_AGE"] = "false"  # if we are using a local DB, we don't want to validate it (but we should be consistent all the time)
        env["GRYPE_DB_AUTO_UPDATE"] = "false"
        env["GRYPE_DB_CACHE_DIR"] = self.db_root
        if self._env:
            env.update(self._env)
        if override:
            env.update(override)
        return env

    @staticmethod
    def parse(result: str, config: artifact.ScanConfiguration) -> List[artifact.Match]:
        logging.debug("parsing grype results")

        results: List[artifact.Match] = []
        obj = json.loads(result)

        # patch the config with the digest found in the results
        repo = config.image.split(":")[0]
        for digest in utils.dig(obj, "source", "target", "repoDigests", default=[]):
            if digest.startswith(repo + "@"):
                config.image_digest = digest.split("@")[1]
                break

        # patch the config with the db information found
        config.detail["db"] = utils.dig(obj, "descriptor", "db", default={})
        db_location = config.detail["db"].get("location", None)  # type: ignore[union-attr]
        if db_location:
            # we should always interpret results with the same DB if the DB is still there (e.g. to support
            # GHSA to CVE mapping for the latest results) if the DB is not there, we try to fall back to
            # the DB found on the system (which isn't ideal, but is better than nothing)
            logging.debug(
                f"using db location found in results to interpret vulnerability definitions: {db_location!r}",
            )
            utils.grype_db.use(db_location)
        else:
            logging.debug(
                "no db location found in results, using system grype DB (not ideal and may cause issues with date filtering)",
            )
            utils.grype_db.use(None)

        for entry in obj["matches"]:
            # TODO: normalize version here
            pkg = artifact.Package(
                name=entry["artifact"]["name"],
                version=entry["artifact"]["version"],
            )
            vuln_id = entry["vulnerability"]["id"]
            cve_id = vuln_id if vuln_id.startswith("CVE-") else None

            if not cve_id:
                for rv in entry.get("relatedVulnerabilities", []):
                    if rv.get("id", "").startswith("CVE-"):
                        cve_id = rv.get("id")
                        break

            vuln = artifact.Vulnerability(id=vuln_id, cve_id=cve_id)
            match = artifact.Match(
                package=pkg,
                vulnerability=vuln,
                fullentry=entry,
                config=config,
            )
            results.append(match)
        return results

    def capture(self, image: str, tool_input: Optional[str]) -> str:
        i = image if tool_input is None else tool_input
        logging.debug(f"running grype with input={i}")
        return self.run("-o", "json", i)

    def run(self, *args, env=None) -> str:
        cmd = [f"{self.path}/grype", *args]
        if self.profile and self.profile.config_path:
            cmd.append("-c")
            cmd.append(self.profile.config_path)
        return subprocess.check_output(
            cmd,
            env=self.env(override=env),
        ).decode("utf-8")

    @staticmethod
    def parse_package_type(full_entry: Optional[Dict[str, Any]]) -> str:
        if not full_entry:
            return "unknown"
        return str(
            utils.dig(
                full_entry,
                "artifact",
                "type",
                default=utils.dig(full_entry, "package_type", default="unknown"),
            ),
        )


def get_import_checksum(db_import_path: str) -> str:
    if db_import_path.endswith(".tar.gz"):
        return handle_legacy_archive(db_import_path)
    if db_import_path.endswith(".tar.zstd") or db_import_path.endswith(".tar.zst"):
        return handle_zstd_archive(db_import_path)
    raise ValueError(f"unsupported db import path: {db_import_path!r}")


# handle_legacy_archive deals with getting the checksum of the DB from the metadata.json file within the DB archive
def handle_legacy_archive(archive_path: str) -> str:
    with tarfile.open(archive_path, "r:gz") as tar:
        metadata_path = None
        for member in tar.getmembers():
            if member.name.endswith("metadata.json"):
                metadata_path = member.name
                break

        if not metadata_path:
            raise ValueError(
                f"could not find metadata.json in {archive_path!r}",
            )

        extractor = tar.extractfile(metadata_path)
        if extractor:
            with extractor as metadata_file:
                metadata = json.load(metadata_file)

        return metadata["checksum"]


# handle_zstd_archive calculates the checksum of the DB from the vulnerability.db file within the DB archive
def handle_zstd_archive(archive_path: str) -> str:
    with open(archive_path, "rb") as compressed_file:
        dctx = zstd.ZstdDecompressor()
        with dctx.stream_reader(compressed_file) as decompressed_stream:
            buffer = io.BytesIO(decompressed_stream.read())

        with tarfile.open(fileobj=buffer, mode="r:") as tar:
            # locate vulnerability.db
            db_member = None
            for member in tar.getmembers():
                if member.name == "vulnerability.db":
                    db_member = member
                    break

            if not db_member:
                raise ValueError(f"could not find vulnerability.db in {archive_path!r}")

            # stream the vulnerability.db file to the hasher
            hasher = xxhash.xxh64()
            with tar.extractfile(db_member) as db_file:
                if not db_file:
                    raise ValueError(f"could not extract vulnerability.db from {archive_path!r}")

                while chunk := db_file.read(8192):
                    hasher.update(chunk)

            return hasher.hexdigest()



import logging
import os

import requests


def get_latest_release_version(project: str, owner: str = "anchore") -> str:
    headers = {}
    token = os.environ.get("GITHUB_TOKEN")
    if token:
        headers["Authorization"] = "Bearer " + token

    response = requests.get(
        f"https://api.github.com/repos/{owner}/{project}/releases/latest",
        headers=headers,
        timeout=15.0,
    )

    if response.status_code >= 400:
        logging.error(
            f"error while fetching latest {project} version: {response.status_code}: {response.reason} {response.text}",
        )

    response.raise_for_status()

    return response.json()["name"]



from asyncio import Future
from concurrent.futures import ThreadPoolExecutor
from typing import List

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    merge_formatted_text,
    to_formatted_text,
)
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Dimension
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.widgets import Label

from yardstick import artifact
from yardstick.cli.explore.image_labels.cve_provider import CveDescriptions
from yardstick.cli.explore.image_labels.edit_note_dialog import EditNoteDialog
from yardstick.cli.explore.image_labels.label_json_editor_dialog import (
    LabelJsonEditorDialog,
)
from yardstick.cli.explore.image_labels.label_manager import LabelManager
from yardstick.cli.explore.image_labels.label_margin import LabelMargin
from yardstick.cli.explore.image_labels.text_area import TextArea


class Worker:
    def __init__(self):
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.futures = []

    def cancel(self):
        for f in self.futures:
            f.cancel()
        self.futures.clear()

    def submit(self, fn):
        self.futures.append(self.executor.submit(fn))


class FilterToolbar:
    def __init__(
        self,
        filter,  # noqa: A002
        apply_fn,
        accept_fn,
    ) -> None:
        def on_change(buf):
            apply_fn(buf.document.text)

        def on_accept(_) -> bool:
            accept_fn()
            return True

        self.accept_fn = accept_fn

        self.buffer = Buffer(
            multiline=False,
            on_text_changed=on_change,
            accept_handler=on_accept,
        )

        self.control = BufferControl(
            focusable=True,
            buffer=self.buffer,
            key_bindings=self.get_key_bindings(),
        )
        self.container = ConditionalContainer(
            content=VSplit(
                [
                    Label(
                        text="Filter by Package/CVE: ",
                        style="class:filter-title",
                        dont_extend_width=True,
                    ),
                    Window(
                        content=self.control,
                        height=1,
                        style="class:filter-toolbar-text",
                    ),
                ],
            ),
            filter=filter,
        )

    def get_key_bindings(self):
        bindings = KeyBindings()

        @bindings.add("/")
        def _done(_):
            self.accept_fn()

        return bindings

    def __pt_container__(self) -> Container:
        return self.container


class ResultSelectionPane:
    def __init__(
        self,
        label_manager: LabelManager,
        dialog_executor,
        match_setter,
        filter_spec: str = "",
    ) -> None:
        self.label_manager = label_manager
        self.dialog_executor = dialog_executor
        self.match_setter = match_setter
        self.selected_entry = 0
        self.result_filter_active = False
        self.cve_descriptions = CveDescriptions()
        self.worker = Worker()

        @Condition
        def search_filter():
            return self.result_filter_active

        self.search_field = FilterToolbar(
            filter=search_filter,
            apply_fn=self._apply_filter,
            accept_fn=self.focus,
        )

        self.text_area = TextArea(
            text=self.get_result_text(),
            read_only=True,
            focusable=True,
            key_bindings=self._get_key_bindings(),
            style="class:select-match-box",
            height=Dimension(preferred=len(self.entries)),
            width=Dimension(preferred=100),
            cursorline=True,
            wrap_lines=True,
            scrollbar=True,
            line_numbers=True,
            right_margins=[LabelMargin(self.label_manager)],
            # always_hide_cursor=True, # not exposed
        )

        if filter_spec:
            filter_spec = f"(filter: {filter_spec})"

        self.container = HSplit(
            [
                Window(
                    content=FormattedTextControl(
                        text=to_formatted_text(
                            f"Match Results {filter_spec}",
                            style="bold reverse",
                        ),
                        focusable=False,
                    ),
                    style="class:pane-title",
                    height=Dimension.exact(1),
                    cursorline=True,
                    always_hide_cursor=True,
                    wrap_lines=False,
                ),
                self.text_area,
                self.search_field,
            ],
        )

        # load the details pane for the current selection
        self._update()

    @property
    def entries(self):
        entries = self.label_manager.match_select_entries
        if entries and self.selected_entry > len(entries):
            self.selected_entry = len(entries) - 2
        return entries

    def focus(self):
        get_app().layout.focus(self.text_area)

    def get_formatted_cve_details(self) -> AnyFormattedText:
        entry = self.get_selected_entry()
        if not entry:
            return merge_formatted_text(
                [to_formatted_text("no selection", style="italic")],
            )
        if self.cve_descriptions.is_cached(entry.match.vulnerability.id):
            text = self.cve_descriptions.get(entry.match.vulnerability.id)
        else:

            def fetch_in_background():
                if self.get_selected_entry().match.vulnerability.id == entry.match.vulnerability.id:
                    self.cve_descriptions.get(entry.match.vulnerability.id)
                    get_app().invalidate()

            self.worker.submit(fetch_in_background)
            text = "loading..."

        return merge_formatted_text([to_formatted_text(text)])

    def get_formatted_result_details(self) -> AnyFormattedText:
        entry = self.get_selected_entry()
        if not entry:
            return merge_formatted_text(
                [to_formatted_text("no selection", style="italic")],
            )
        return entry.get_formatted_details()

    def get_result_text(self) -> AnyFormattedText:
        result: List[AnyFormattedText] = []

        if not self.entries:
            return "no selection"

        for _, entry in enumerate(self.entries):
            result.append(entry.display)

        return "\n".join(result)  # type: ignore[arg-type]

    def get_selected_entry(self):
        if self.selected_entry >= len(self.entries):
            return None
        return self.entries[self.selected_entry]

    def _update(self):
        entry = self.get_selected_entry()
        if entry:
            self.match_setter(entry.match)
            self.worker.cancel()

    def _toggle_filter(self):
        self.result_filter_active = not self.result_filter_active
        if self.result_filter_active:
            get_app().layout.focus(self.search_field)
        else:
            self._apply_filter("")

    def _apply_filter(self, text):
        self.label_manager.apply_filter(text)
        self.text_area.text = self.get_result_text()
        self.selected_entry = 0
        self._update()
        get_app().invalidate()

    def _get_key_bindings(self) -> KeyBindings:  # noqa: C901, PLR0915
        kb = KeyBindings()

        @kb.add("up")
        def _go_up(event) -> None:
            if len(self.entries):
                selected_entry = self.selected_entry - 1
                if selected_entry >= 0:
                    self.selected_entry = selected_entry
                    self.text_area.control.move_cursor_up()
                    self._update()

        @kb.add("down")
        def _go_down(event) -> None:
            if len(self.entries):
                selected_entry = self.selected_entry + 1
                if selected_entry < len(self.entries):
                    self.selected_entry = selected_entry
                    self.text_area.control.move_cursor_down()
                    self._update()

        @kb.add("f")
        def _fp(event) -> None:
            entry = self.get_selected_entry()
            if entry:
                self.label_manager.add_label_entry(
                    entry.match,
                    artifact.Label.FalsePositive,
                )
            self._update()
            get_app().invalidate()

        @kb.add("t")
        def _tp(event) -> None:
            entry = self.get_selected_entry()
            if entry:
                self.label_manager.add_label_entry(
                    entry.match,
                    artifact.Label.TruePositive,
                )
            self._update()
            get_app().invalidate()

        @kb.add("?")
        def _unknown(event) -> None:
            entry = self.get_selected_entry()
            if entry:
                self.label_manager.add_label_entry(entry.match, artifact.Label.Unclear)
            self._update()
            get_app().invalidate()

        @kb.add("/")
        def _filter(event) -> None:
            self._toggle_filter()

        def get_single_label_entry():
            selected_entry = self.get_selected_entry()
            if not selected_entry:
                return None
            label_entries = self.label_manager.get_label_entries_by_match(
                selected_entry.match,
            )

            if not label_entries:
                return None

            return label_entries[0]

        @kb.add("e")
        def _edit_note(event) -> None:
            entry = get_single_label_entry()

            if not entry:
                return

            def done(fut: Future):
                if fut.done() and fut.result() is not None:
                    note = fut.result()
                    self.label_manager.edit_label_entry_note(entry.ID, note)
                    self._update()

            self.dialog_executor(
                dialog=EditNoteDialog(
                    title="Edit note",
                    label_text="Note:",
                    text_area=True,
                    ok_button_text="Update",
                    default_value=entry.note,
                ),
                done_callback=done,
            )

        @kb.add("j")
        def _edit_json(event) -> None:
            entry = get_single_label_entry()

            def done(fut: Future):
                if fut.done() and fut.result() is not None:
                    self.label_manager.edit_label_entry_json(entry.ID, fut.result())
                    self._update()

            self.dialog_executor(
                dialog=LabelJsonEditorDialog(entry),
                done_callback=done,
            )

        @kb.add("backspace")
        @kb.add("delete")
        def _delete_label(event) -> None:
            entry = get_single_label_entry()

            self.label_manager.remove_label_entry(entry.ID)

            self._update()

        return kb

    def __pt_container__(self) -> Container:
        return self.container
